#ifndef __IOCTL_BT848_H__
#define __IOCTL_BT848_H__
#define MAJOR_NUM	81
typedef struct _DMA_BUFFER_INFO
{
        int blocks;
        int rings;
        int currentWrite;
        int int_stat;
        int int_mask;
} DMA_BUFFER_INFO, *P_DMA_BUFFER_INFO;

typedef struct {
        long LNB1;
        long LNB2;
} LNB_LOF, *P_LNB_LOF;

typedef enum {
    true = 1, 
    false = 0
} tBool;

typedef struct _DISEQC_COMMAND
{
    unsigned char BitCount;
    unsigned char Frame;
    unsigned char Address;
    unsigned char Command;
    unsigned char Data1;
    unsigned char Data2;
} DISEQC_COMMAND, *P_DISEQC_COMMAND;

typedef struct _LNB_INFO
{
    tBool  Power_Enable;
    tBool  K22_Enable;
    tBool  Tone_Burst_Enable;
    tBool  Data_Burst_Enable;
} LNB_INFO, *P_LNB_INFO;

typedef char *P_CHAR;
typedef unsigned char STRING[258];

typedef char PROTOCOL[8];
typedef PROTOCOL *P_PROTOCOL;

typedef struct
{
    int  pid;
    int  if_num;
}dst_net_if;

typedef struct _READ_BUF
{
    void *pBuf;
    unsigned long Len;
    unsigned long Pos;
} READ_BUF, *P_READ_BUF;

typedef struct _DEMULTIPLEXER_VALUE
{
    unsigned long   PID;  //for TH_PID_TS_PACKET, TH_PID_PES_PACKET, TH_ALL_PID_TS_PACKET
    unsigned long   vPID; //for TH_PID_PS_PACKET
	unsigned long   aPID; //for TH_PID_PS_PACKET
	int             StreamType;	
	char            *FileName;  //not used in driver
} DEMULTIPLEXER_VALUE, *P_DEMULTIPLEXER_VALUE;

typedef struct _TRANSORT_PACKET_FORMAT
{
    unsigned char  pSession[204];  //Session Data (video | audio)
    int            SessionLen;    //Session len
    unsigned long  TSSize;    //188 or 204   
    
    //Fields
    int       transport_error_indicator;
    int       payload_unit_start_indicator;
    int       pid;
    int       transport_scrambling_control;
    int       adaptation_field_control;
    int       adaptation_field_length;
    int       continuity_counter;

} TRANSORT_PACKET_FORMAT, *P_TRANSORT_PACKET_FORMAT ;
    
typedef struct _READ_STRU
{
    unsigned long       ReadPosition;
    int                 StartBitEnable;
    int                 AStartBitEnable;
    int                 VStartBitEnable;
    unsigned char	    TempBuffer[204];
    unsigned long   	TempBufferPosition;
    DEMULTIPLEXER_VALUE DMV;
    TRANSORT_PACKET_FORMAT TPFData;	    
    READ_BUF ReadBuf;
    int  bRead_Enable;
}   READ_STRU, *P_READ_STRU; 
READ_STRU read_stru[NBKTR];    
    
typedef struct _DISEQC_VALUE
{
    int nDiseqc_Port; //port A: 1, port B:2,  port C:3,  port D:4
    int nLnb_Type;
    int nPolarity;
} DISEQC_VALUE, *P_DISEQC_VALUE;

typedef struct _SIGNAL_VALUE
{
    int nLockFlag; 
    int nDevSigQuality;
    int nDevSigStrength;
} SIGNAL_VALUE, *P_SIGNAL_VALUE;

typedef struct _CARD_VALUE
{
    char sCard_Name[8]; 
    int nTS_SIZE;
} CARD_VALUE, *P_CARD_VALUE;
    
#define READ_BUF_SIZE           1*1024*1024 //1M
#define ReadBuf_PAGE_ORDER      8  //1 page = 4k
#define READ_BUF_OFFSET_SIZE    300*1024 //300k

//Stream type
#define TH_PID_TS_PACKET        0  //One PID's TS
#define TH_PID_PES_PACKET       1  //One PID's PES
//#define TH_PID_PS_PACKET        2  //Merge Video & Auido => Program Stream
#define TH_ALL_PID_TS_PACKET    3  //ALL PID's TS
#define TH_VA_TS_PACKET         4  //Mix Video and Audio PID's TS

#define DMABUFFERLEN_2          255*3*4092UL

#define LNB_TYPE_NORMAL     0  //freq=5150
#define LNB_TYPE_UNIVERSAL  1  //freq=10600
#define LNB_TYPE_CUSTOM     2  //freq=11300

#define	TUNER_SETTUNER          _IOW(MAJOR_NUM, 0, P_PROGRAM_VALUE)
#define	TUNER_GETTUNER          _IOR(MAJOR_NUM, 1, P_PROGRAM_VALUE)
#define	TUNER_STOP              _IOW(MAJOR_NUM, 2, int)
#define DST_INIT		        _IOW(MAJOR_NUM, 10, P_LNB_LOF)
#define DST_GET_STATUS		    _IOR(MAJOR_NUM, 11, P_PROGRAM_VALUE)
#define DST_START_CAPTURE	    _IOW(MAJOR_NUM, 12, int)
#define DST_STOP_CAPTURE	    _IOW(MAJOR_NUM, 13, int)
#define DST_GET_BUFFER_INFO	    _IOR(MAJOR_NUM, 14, P_DMA_BUFFER_INFO)

#define DST_PUT_COMMAND	        _IOW(MAJOR_NUM, 15, P_PROTOCOL)
#define DST_GET_REPLY	        _IOR(MAJOR_NUM, 16, P_PROTOCOL)

#define READ_CAMSG              _IOR(MAJOR_NUM, 20, STRING *)
#define WRITE_CAMSG             _IOW(MAJOR_NUM, 21, STRING *)
#define RESET_MCU               _IOW(MAJOR_NUM, 22, int)

#define NET_ADD_IF              _IOWR(MAJOR_NUM, 25, struct dst_net_if *)
#define NET_REMOVE_IF           _IOW(MAJOR_NUM,  26, uint16_t)

#define	INIT_DEMUX              _IOW(MAJOR_NUM, 30, P_DEMULTIPLEXER_VALUE)
#define	DST_SET_VOLTAGE         _IOW(MAJOR_NUM, 31, int)
#define	DST_SET_DISEQC          _IOW(MAJOR_NUM, 32, P_DISEQC_VALUE)
#define	DST_GET_SIGNAL          _IOWR(MAJOR_NUM, 33, P_SIGNAL_VALUE)
#define	DST_GET_CARD_TYPE       _IOR(MAJOR_NUM, 34, P_CARD_VALUE)
#define	DST_SET_LNB             _IOW(MAJOR_NUM, 35, P_LNB_INFO)
#define	DST_SET_DISEQC_COMMAND  _IOW(MAJOR_NUM, 36, P_DISEQC_COMMAND)


#endif

