#ifdef linux
extern unsigned long loops_per_sec;
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/mm.h>
#include "twinhan_linux.h"
#endif

#ifndef LIBC6
#include "/usr/include/errno.h"
//#include "/usr/include/signal.h"
#include <asm/signal.h>
#endif

#define SUCCESS 0
#include "ioctl_twinhan.h"
#include "brktree_reg.h"
#include "twinhan_common.h"
#include "dstnet.h"
#define RETRIES_COUNT   3
STRING ca_input, ca_output;
#define HAS_MSG 1
#define ERROR_STATUS -1
#define NO_MSG  0
int ca_status = NO_MSG;
static void	build_dma_prog( bktr_ptr_t bktr);
bktr_reg_t brooktree[ NBKTR ];
int brooktree_cnt;

static int Lock_Tuner(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue);
static int Lock_Tuner_Program(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue, P_TUNER_DATA tunerData);
static int New_Lock_Tuner_Program(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue, P_NEW_TUNER_DATA tunerData);

static unsigned char new_tunerDataCheckSumOK(P_NEW_TUNER_DATA tunerData);
static void new_convertProgram(P_PROGRAM_VALUE programValue, unsigned char * tunerData, P_LNB_LOF pLNB, int Flag22K);
static void pulldownGPIO(bktr_ptr_t bktr);
static void releaseGPIO(bktr_ptr_t bktr);
static unsigned isGPIOlow(bktr_ptr_t bktr);
#define BKTR_FM1      0x6
#define BKTR_FM3      0xe
#define BKTR_VRE      0x4
#define BKTR_VRO      0xC
#define BKTR_PXV      0x0
#define BKTR_EOL      0x1
#define BKTR_SOL      0x2
#define OP_WRITE      (0x1 << 28)
#define OP_SKIP       (0x2 << 28)
#define OP_WRITEC     (0x5 << 28)
#define OP_JUMP	      (0x7 << 28)
#define OP_SYNC	      (0x8 << 28)
#define OP_WRITE123   (0x9 << 28)
#define OP_WRITES123  (0xb << 28)
#define OP_SOL	      (1 << 27)
#define OP_EOL	      (1 << 26)
#define ALP_LEN		4092
#define AFP_LEN		255
#define DMA_BUFFER_BLOCK   3 
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

static dst_net_t  *dstnet;
static unsigned char IP_mac[6];   //*****
static int  IP_pid;

extern unsigned char *mmbuf;  //defined by dambuffer.c
static int TS_SIZE;
static unsigned char  tsbuf[204];
static int IP_free_flag;
typedef struct
{
    uint8_t   *pbuf;
    size_t    pos ;
    size_t    offset;

    int       transport_error_indicator;
    int       payload_unit_start_indicator;
    int       pid;
    int       transport_scrambling_control;
    int       adaptation_field_control;
    int       continuity_counter;

    uint8_t   section[4096];
    int       section_size;
    int       section_valid;
}demuxer_t ;
static demuxer_t  *demuxer;

static int section_thread( void *);

static int dst_net_add_if(u16 pid);
static int dst_net_remove_if(int num) ;
static void interrupt8820(bktr_ptr_t bktr);
static void reset8820(bktr_ptr_t bktr);
static int i2cWriteMCU( bktr_ptr_t bktr, P_TUNER_DATA tunerData );

//struct task_struct  *section_thread;

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void ddelay(int tms)
{
     current->state = TASK_INTERRUPTIBLE;
     schedule_timeout(HZ*tms/100);
}

static void build_dma_prog( bktr_ptr_t bktr )
{
    DBGMSG(">>build_dma_prog()\n");
    
    bt848f1_ptr_t		bt848f1;
    volatile u_long		target_buffer, buffer, target;
    volatile  u_long	*dma_prog;
    int i,j;
    bt848f1 = bktr->f1base;
    bt848f1->int_mask = ALL_INTS_DISABLED;
    bt848f1->gpio_dma_ctl &= ~FIFO_RISC_ENABLED;
    bt848f1->risc_strt_add = vtophys(bktr->dma_prog);
    bt848f1->audio_packet_length = (AFP_LEN << 16) | ALP_LEN;
    dma_prog = (u_long *) bktr->dma_prog;
    target_buffer = (u_long) vtophys(bktr->bigbuf);
    buffer = target_buffer;
    target = target_buffer;
    for (i = 0; i < DMA_BUFFER_BLOCK; i++)  {
        *dma_prog++ = OP_SYNC  | 1 << 15 | BKTR_FM1;
        *dma_prog++ = 0;
        for (j=0; j<AFP_LEN; j++) {
    		*dma_prog++ = OP_WRITE | OP_SOL | OP_EOL | ALP_LEN;
    		*dma_prog++ = target;
    		target += ALP_LEN;
		}
		*dma_prog++ = OP_SYNC | BKTR_VRO;
        *dma_prog++ = 0;
    }
   *dma_prog++ = OP_JUMP ;
   *dma_prog++ = (u_long ) vtophys(bktr->dma_prog);
   
   DBGMSG("<<build_dma_prog()\n");
}

#define DA_SBR 		(1 << 14)
#define DA_APP  	(1 << 5)
#define DA_IOM0 	(1 << 6)
#define DA_DPM  	(1 << 15)
#define ACAP_EN		(1 << 4)
#define RISC_ENABLE 	(1 << 1)
#define FIFO_ENABLE 	(1 << 0)
#define PKTP            (2 << 2)
#define SCERR           (1 << 19)
#define OCERR           (1 << 18)
#define PABORT          (1 << 17)
#define RIPERR          (1 << 16)
#define PPERR           (1 << 15)
#define FDSR            (1 << 14)
#define FTRGT           (1 << 13)
#define FBUS            (1 << 12)
#define RISCI           (1 << 11)
#define GPINT           (1 << 9)
#define VPRES           (1 << 5)
#define HLOCK           (1 << 4)
#define OFLOW           (1 << 3)
#define HSYNC           (1 << 2)
#define VSYNC           (1 << 1)
#define FMTCHG          (1 << 0)

void start_capture( bktr_ptr_t bktr)
{
    DBGMSG(">>start_capture()\n");

	bt848f1_ptr_t	bt848f1;
	bt848f1 = bktr->f1base;
	bt848f1->int_stat = bt848f1->int_stat;
    bt848f1->gpio_dma_ctl = 0x02204f0c;
	bktr->flags |= DST_CAPTURING ;
    bktr->rings = 0;
	build_dma_prog(bktr);
	bktr->dma_prog_loaded = TRUE;
	bt848f1->risc_strt_add = vtophys(bktr->dma_prog);
    bt848f1->int_mask |= SCERR | OCERR | FDSR | FTRGT | FBUS | RISCI ;
    bt848f1->gpio_dma_ctl |= (FIFO_ENABLE | ACAP_EN | DA_APP | RISC_ENABLE);
    DBGMSG("<<start_capture()\n");
}

void stop_capture( bktr_ptr_t bktr)
{
    DBGMSG(">>stop_capture()\n");
        
	bt848f1_ptr_t bt848;
	bt848 = bktr->f1base;
	bt848->int_stat = bt848->int_stat;
	bktr->flags &= ~DST_CAPTURING ;
    bt848->gpio_dma_ctl &= ~(FIFO_ENABLE | RISC_ENABLE | ACAP_EN | DA_APP);
        
    DBGMSG("<<stop_capture()\n");
}

int getCaptureStatus( bktr_ptr_t bktr)
{
	return bktr->flags;
}

#define I2CBITTIME      0x80
#define I2C_READ	0x01
#define I2C_COMMAND	(I2CBITTIME | BT848_DATA_CTL_I2CSCL | BT848_DATA_CTL_I2CSDA)
#define MCU_RDADDR	0xab
#define MCU_WRADDR	0xaa
#define I2CMODE		(1 << 7)
#define I2CRATE		(1 << 6)
#define I2CNOSTOP	(1 << 5)
#define I2CNOS1B	(1 << 4)
#define I2CW3BRA	(1 << 2)
#define I2CMODE		(1 << 7)
#define I2CRACK         (1 << 25)

static volatile int I2cDone = 0;
static unsigned char tunerSetCheckSum(P_TUNER_DATA tunerData);
#define ACK_TIMEOUT 100

static int i2cWriteMCU( bktr_ptr_t bktr, P_TUNER_DATA tunerData )
{
    DBGMSG(">>i2cWriteMCU()\n");
    
    u_long		data;
    bt848_ptr_t	bt848;
    int i;
    unsigned char *pData;
    int retries = 5;
    int timeout;
    pData = (unsigned char *)tunerData;
    bt848 = bktr->base;
    while ( -- retries ) {
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        bt848->int_stat = 0xffffffff;
        bt848->int_mask =  (BT848_INT_RACK | BT848_INT_I2CDONE);
        data = (tunerData->address << 24) |  I2C_COMMAND;
        data |= (pData[1]) << 16;
        data &= ~I2CNOS1B;
        data |=	I2CNOSTOP;
        data &= ~I2CW3BRA;
        I2cDone = 0;
        bt848->i2c_data_ctl = data;
        timeout = ACK_TIMEOUT;
        while (! I2cDone && timeout --) {
            udelay(1000);
        }
        if (timeout <= 0) {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        if ( (bt848->int_stat & I2CRACK) == 0 ) {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        I2cDone = 0;
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        pData = (unsigned char *)tunerData;
        for (i=2; i<8; i++){
            bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
            data = (pData[i] << 24) | (I2CNOS1B | I2CNOSTOP | I2C_COMMAND);
            I2cDone = 0;
            bt848->i2c_data_ctl = data;
            bt848->int_mask |=   BT848_INT_I2CDONE;
            timeout = ACK_TIMEOUT;
            while (! I2cDone && timeout --){ 
				udelay(1000);				
			}
            if (timeout <= 0) {
                releaseGPIO(bktr);
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout(HZ*10/200);
                pulldownGPIO(bktr);
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout(HZ*10/200);
                continue;
            }
            if ( (bt848->int_stat & I2CRACK) == 0 )  
                break;
        }
        if (i < 8){
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            continue;
        }
        data = (pData[8] << 24) | (I2CNOS1B | I2C_COMMAND);
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        I2cDone = 0;
        bt848->i2c_data_ctl = data;
        bt848->int_mask |=  (BT848_INT_RACK | BT848_INT_I2CDONE);
        timeout = ACK_TIMEOUT;
        while (! I2cDone && timeout --) { 
			udelay(1000);			
		}
        if (timeout <= 0) { 
             releaseGPIO(bktr);
             current->state = TASK_INTERRUPTIBLE;
             schedule_timeout(HZ*10/200);
             pulldownGPIO(bktr);
             current->state = TASK_INTERRUPTIBLE;
             schedule_timeout(HZ*10/200);
             continue;
        }
        break;
    }
      
    DBGMSG("<<i2cWriteMCU()\n");
    if (retries > 0)  
        return 0 ;
    return 1;
}

static int i2cReadVerify( bktr_ptr_t bktr, unsigned char *verify)
{
    DBGMSG(">>i2cReadVerify()\n");
    
    u_long          data;
    bt848_ptr_t	bt848;
    int timeout;
    bt848 = bktr->base;
    bt848->int_stat =  BT848_INT_I2CDONE;
    data = (MCU_RDADDR << 24) | I2C_COMMAND;
    data &= ~I2CNOSTOP;
    data &= ~I2CNOS1B;
    data &= ~I2CW3BRA;
    bt848->i2c_data_ctl = data ;
    I2cDone = 0;
    bt848->int_mask |=  BT848_INT_I2CDONE;
    timeout = ACK_TIMEOUT;
    while (! I2cDone && timeout --)	{ 
		udelay(1000);
	}
    if (timeout <= 0) {
        *verify = 0;
        return 1;
    }
    *verify = (bt848->i2c_data_ctl >> 8) & 0xff;
        
    DBGMSG("<<i2cReadVerify()\n");
    return SUCCESS;
}

static int i2cReadTuner( bktr_ptr_t bktr, P_TUNER_DATA tunerdata )
{
    DBGMSG(">>i2cReadTuner()\n");
        
	bt848_ptr_t	bt848;
	u_long 		data;
	int 		i;
    int             timeout;
	unsigned char *pByte = (unsigned char *)tunerdata;
	bt848 = bktr->base;
	bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
	tunerdata->address = MCU_RDADDR;
	for (i=1; i<9; i++){
        data = (MCU_RDADDR << 24) | I2C_COMMAND | I2CNOS1B | I2CNOSTOP | I2CW3BRA;
        if (i == 1)     data &= ~I2CNOS1B;
        if (i == 8)	data &= ~(I2CW3BRA | I2CNOSTOP);
        I2cDone = 0;
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        bt848->i2c_data_ctl = data;
        bt848->int_mask |=   BT848_INT_I2CDONE;
        timeout = ACK_TIMEOUT;
        while (! I2cDone && timeout --) { 
			udelay(1000);
		}
        if (timeout <= 0) {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            return 1;
        }
	 	pByte[i] = ( (bt848->i2c_data_ctl >> 8) & 0xff );
	}
	
	DBGMSG("<<i2cReadTuner()\n");
	return 0;
}

static int new_i2cWriteMCU( bktr_ptr_t bktr, u_char *writeData )
{
    DBGMSG(">>new_i2cWriteMCU()\n");
    
    u_long		data;
    bt848_ptr_t	bt848;
    int i;
    unsigned char *pData;
    int retries = 5;
    int timeout;
    int length;
    pData = (unsigned char *)writeData;
    length = pData[0];
    if (length == 0)
        length = 7;
    bt848 = bktr->base;
    while ( -- retries )   {
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        bt848->int_stat = 0xffffffff;
        bt848->int_mask =  (BT848_INT_RACK | BT848_INT_I2CDONE);
        data = (MCU_WRADDR << 24) |  I2C_COMMAND;
        data |= (pData[0]) << 16;
        data &= ~I2CNOS1B;
        data |=	I2CNOSTOP;
        data &= ~I2CW3BRA;
        I2cDone = 0;
        bt848->i2c_data_ctl = data;
        timeout = ACK_TIMEOUT;
        while (! I2cDone && timeout --) { 
			udelay(1000); // 1ms
		}
        if (timeout <= 0) {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        if ( (bt848->int_stat & I2CRACK) == 0 ){
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        I2cDone = 0;
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        for (i=1; i< length; i++) {
            bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
            data = (pData[i] << 24) | (I2CNOS1B | I2CNOSTOP | I2C_COMMAND);
            I2cDone = 0;
            bt848->i2c_data_ctl = data;
            bt848->int_mask |=   BT848_INT_I2CDONE;
            timeout = ACK_TIMEOUT;
            while (! I2cDone && timeout --) 
            {
				udelay(1000);
			}
            if (timeout <= 0) {
                releaseGPIO(bktr);
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout(HZ*10/200);
                pulldownGPIO(bktr);
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout(HZ*10/200);
                continue;
            }
            if ( (bt848->int_stat & I2CRACK) == 0 ) break;
        }
        if (i< length){
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            continue;
        }
        data = (pData[ length ] << 24) | (I2CNOS1B | I2C_COMMAND);
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
        I2cDone = 0;
        bt848->i2c_data_ctl = data;
        bt848->int_mask |=  (BT848_INT_RACK | BT848_INT_I2CDONE);
        timeout = ACK_TIMEOUT;
        while (! I2cDone && timeout --) 
        {
			udelay(1000);			
		}
        if (timeout <= 0){
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            continue;
        }
		break;
    }
        
    DBGMSG("<<new_i2cWriteMCU()\n");
    
    if (retries > 0)    
        return( 0 );                
    return 1;
}

static int new_i2cReadVerify( bktr_ptr_t bktr, unsigned char *verify)
{
    DBGMSG(">>new_i2cReadVerify()\n");
    
    u_long    data;
    bt848_ptr_t	bt848;
    int timeout;
    bt848 = bktr->base;
    bt848->int_stat =  BT848_INT_I2CDONE;
    data = (MCU_RDADDR << 24) | I2C_COMMAND;
    data &= ~I2CNOSTOP;
    data &= ~I2CNOS1B;
    data &= ~I2CW3BRA;
    bt848->i2c_data_ctl = data ;
    I2cDone = 0;
    bt848->int_mask |=  BT848_INT_I2CDONE;
    timeout = ACK_TIMEOUT;
    while (! I2cDone && timeout --) 
    {
		udelay(1000);
	}
    if (timeout <= 0){
        *verify = 0;
        return 1;
    }
    *verify = (bt848->i2c_data_ctl >> 8) & 0xff;
        
    DBGMSG("<<new_i2cReadVerify()\n");
    return SUCCESS;
}

static int new_i2cReadTuner( bktr_ptr_t bktr, unsigned char * readData )
{
    DBGMSG(">>new_i2cReadTuner()\n");
    
    bt848_ptr_t	bt848;
    u_long 		data;
    int 		i;
    int             timeout;
    unsigned char *pByte = (unsigned char *)readData;
    int length;
    bt848 = bktr->base;
    bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
    length = 255;
	for (i=0; i<= length; i++){
		data = (MCU_RDADDR << 24) | I2C_COMMAND | I2CNOS1B | I2CNOSTOP | I2CW3BRA;
        if (i == 0)      
            data &= ~I2CNOS1B;
		if (i== length ) 
		    data &= ~(I2CW3BRA | I2CNOSTOP);
        I2cDone = 0;
        bt848->int_stat = (BT848_INT_RACK | BT848_INT_I2CDONE);
		bt848->i2c_data_ctl = data;
        bt848->int_mask |=   BT848_INT_I2CDONE;
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
        timeout = ACK_TIMEOUT;        
        while (!I2cDone && timeout--)
		{
			udelay(1000);
		}
        
        if (timeout <= 0)   {
             releaseGPIO(bktr);
             current->state = TASK_INTERRUPTIBLE;
             schedule_timeout(HZ*10/200);
             pulldownGPIO(bktr);
             current->state = TASK_INTERRUPTIBLE;
             schedule_timeout(HZ*10/200);
             return 1;
        }
        if ((bt848->int_stat & I2CRACK) == 0) {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            pulldownGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/200);
            continue;
        }
	 	pByte[i] = ((bt848->i2c_data_ctl >> 8) & 0xff);
        if (i==0){
             length = pByte[0];
             if (length == 0)  
                length = 7;
        }
	}
	
	DBGMSG("<<new_i2cReadTuner()\n");
	return 0;
}

u_long frequencyRef[16]={2,4,8,16,32,64,128,256,24,5,10,20,40,80, 160, 320};

static void convertFrequency(unsigned long  frequency,P_LNB_LOF pLNB,TUNER_DATA *tunerData)
{
        DBGMSG(">>convertFrequency()\n");
        
	u_long tunerStep;
        long InFrequency;
        long InNumber;
        int divRate,i;
	tunerStep=1000000;
	InFrequency=( frequency> pLNB->LNB1 ) ? (frequency-pLNB->LNB2):(pLNB->LNB1-frequency);
	if(InFrequency <1531)		tunerData->flag |= 0x0004;
	else	tunerData->flag &= 0xfffb;
	InNumber=InFrequency*1000/(tunerStep/1000);
	tunerData->frequencyMSB=(InNumber&0x00007f00)>>8;
	tunerData->frequencyLSB=InNumber& 0x000000ff;
	divRate=4000000L/tunerStep;
	for(i=0;i<16;i++)
	if(frequencyRef[i]==divRate) break;
	if(i>15) i=15;
	tunerData->tunerStep=i;
	
	DBGMSG("<<convertFrequency()\n");
}

static long  BinaryFloatDiv(long n1, long n2, int precision)
{
    DBGMSG(">>BinaryFloatDiv()\n");
    
    int i=0;
    long result=0;
    while(i<=precision) {
        if (n1<n2) {
        	result*=2;
        	n1*=2;
        } else {
        	result=result*2+1;
        	n1=(n1-n2)*2;
        }
        i++;
    }
    
    DBGMSG("<<BinaryFloatDiv()\n");
    return result;
}

static void convertSymbolRate(long SymbolRate,P_TUNER_DATA tunerData)
{
    DBGMSG(">>convertSymbolRate()\n");
        
    long Result = BinaryFloatDiv(SymbolRate*1000,88000000,20);
    tunerData->symbolRateHSB =(Result>>12)&0xFF;
    tunerData->symbolRateMSB =(Result>>4)&0xFF;
    tunerData->symbolRateLSB =((Result&0x0F)<<4)&0xF0;
    
    DBGMSG("<<convertSymbolRate()\n");
}

static void setFlag1318V(int flag1318V, P_TUNER_DATA tunerData)
{
     DBGMSG(">>setFlag1318V()\n");
        
     if (flag1318V) 
        tunerData->flag |= 0x40;
     else  
        tunerData->flag &= 0xbf;
     
     DBGMSG("<<setFlag1318V()\n");
}

static void setFlag22K(int flag22K, P_TUNER_DATA tunerData)
{
     DBGMSG(">>setFlag22K()\n");   
     
     tunerData->flag &= 0xfc;
     if (flag22K)   
        tunerData->flag |= 0x01;
     
     DBGMSG("<<setFlag22K()\n");   
}

static unsigned char tunerSetCheckSum(P_TUNER_DATA tunerData)
{
    DBGMSG(">>tunerSetCheckSum()\n");   
        
	unsigned char *pByte = (unsigned char *) tunerData;
	unsigned char chksum = 0;
    int i;
    
	for (i=1; i <8; i++)
	    chksum += pByte[i];
	chksum = ~chksum+1;
	pByte[8] = chksum;
	
	DBGMSG("<<tunerSetCheckSum()\n");   
	return chksum;
}

static unsigned char tunerDataCheckSumOK(P_TUNER_DATA tunerData)
{
    DBGMSG(">>tunerDataCheckSumOK()\n"); 
          
	unsigned char *pByte = (unsigned char *) tunerData;
	unsigned char chksum = 0;
    int i;
   	for ( i=1; i <8; i++)
   		chksum += pByte[i];
	chksum = ~chksum+1;
	
	DBGMSG("<<tunerDataCheckSumOK()\n"); 
	return ( chksum == pByte[8]);
}

static void convertProgram(P_PROGRAM_VALUE programValue, P_TUNER_DATA tunerData, P_LNB_LOF pLNB, int Flag22K)
{
    DBGMSG(">>convertProgram()\n"); 
    
    tunerData->flag = 0;
	convertFrequency(programValue->frequency, pLNB, tunerData);
	convertSymbolRate(programValue->symbolRate, tunerData);
	if (programValue->i_Q)
	    tunerData->flag |= 0x80;
	if (programValue->symbolRate>8000) 
	    tunerData->flag |= 0x20;
    setFlag1318V(programValue->h_V, tunerData);
    setFlag22K(Flag22K, tunerData);
	tunerSetCheckSum(tunerData);
	
	DBGMSG("<<convertProgram()\n"); 
}

static void pulldownGPIO(bktr_ptr_t bktr)
{
    DBGMSG(">>pulldownGPIO()\n"); 
        
	bt848_ptr_t	bt848;
	bt848 = bktr->base;
	bt848->gpio_out_en = 0x00000001;
	bt848->gpio_data = 0;
	
	DBGMSG("<<pulldownGPIO()\n"); 
}

static void releaseGPIO(bktr_ptr_t bktr)
{
    DBGMSG(">>releaseGPIO()\n"); 
        
	bt848_ptr_t	bt848;
	bt848 = bktr->base;
	bt848->gpio_out_en = 0x00000000;
	
	DBGMSG("<<releaseGPIO()\n"); 
}

static unsigned isGPIOlow(bktr_ptr_t bktr)
{
    DBGMSG(">>isGPIOlow()\n"); 
        
	bt848_ptr_t bt848;
	bt848 = bktr->base;
	
	DBGMSG("<<isGPIOlow()\n"); 
	return ( (bt848->gpio_data & 0x00000001) == 0);
}	

static void inter8820(bktr_ptr_t bktr)
{
    DBGMSG(">>inter8820()\n"); 
        
	bt848_ptr_t	bt848;
	bt848 = bktr->base;
	bt848->gpio_out_en |= 0x00000002;
	bt848->gpio_data |= 0x02;
	
	DBGMSG("<<inter8820()\n"); 
}

static void release8820(bktr_ptr_t bktr)
{
    DBGMSG(">>release8820()\n"); 
        
	bt848_ptr_t	bt848;
	bt848 = bktr->base;
	bt848->gpio_out_en |= 0x00000002;
	bt848->gpio_data &= ~0x0002;
	
	DBGMSG("<<release8820()\n"); 
}

static void  convertTunerData(TUNER_DATA *tunerData,P_PROGRAM_VALUE programValue, P_LNB_LOF pLNB, u_long frequency)
{
    DBGMSG(">>convertTunerData()\n"); 
        
	long Ltmp,Ltmp2;
	long Mclk;
	long number;
	int tunerStep;
	
	Mclk =	88000000/ 4096L;
	Ltmp = ((tunerData->symbolRateHSB<<12)+(tunerData->symbolRateMSB<<4))/16;
	Ltmp *= Mclk;
	Ltmp /=16;
	Ltmp2=(tunerData->symbolRateLSB*Mclk)/256;
	Ltmp += Ltmp2;
	programValue->symbolRate = Ltmp;
	number =((tunerData->frequencyMSB&0x007f)<<8) + tunerData->frequencyLSB;
    tunerStep = 1000000;
	programValue->frequency= (number*tunerStep)/1000000;
    if (frequency < pLNB->LNB1)       
        programValue->frequency = pLNB->LNB1 - programValue->frequency;
    else
        programValue->frequency = programValue->frequency + pLNB->LNB2;
               
    DBGMSG("<<convertTunerData()\n"); 
}

static void  new_convertTunerData(NEW_TUNER_DATA *tunerData,P_PROGRAM_VALUE programValue, P_LNB_LOF pLNB, u_long frequency)
{
    DBGMSG(">>new_convertTunerData()\n"); 
        
	long Ltmp,Ltmp2;
	long Mclk;
	long number;
	int tunerStep;
	Mclk =	88000000/ 4096L;
	Ltmp = ((tunerData->symbolRateHSB<<12)+(tunerData->symbolRateMSB<<4))/16;
	Ltmp *= Mclk;
	Ltmp /=16;
	Ltmp2=(tunerData->symbolRateLSB*Mclk)/256;
	Ltmp += Ltmp2;
	programValue->symbolRate = Ltmp;
	number =((tunerData->frequencyMSB&0x007f)<<8) + tunerData->frequencyLSB;
    tunerStep = 1000000;
	programValue->frequency= (number*tunerStep)/1000000;
    if (frequency < pLNB->LNB1)   
        programValue->frequency = pLNB->LNB1 - programValue->frequency;
    else      
        programValue->frequency = programValue->frequency + pLNB->LNB2;
        
    DBGMSG("<<new_convertTunerData()\n"); 
}

static int stop_tuner(bktr_ptr_t bktr)
{
    DBGMSG(">>stop_tuner()\n"); 
    DBGMSG("<<stop_tuner()\n"); 
    return SUCCESS;
}	

static int Lock_Tuner(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue)
{
    DBGMSG(">>Lock_Tuner(): programValue->frequency=%ld, programValue->symbolRate=%ld\n", programValue->frequency, programValue->symbolRate); 
    DBGMSG(">>Lock_Tuner(): bktr->card_type=%d \n", bktr->card_type);
    int tmpfrequency = 0;
    TUNER_DATA tunerData;
    NEW_TUNER_DATA new_tunerData;
    
    switch (bktr->card_type) {
    case 1:
    case 2:
        {
            if (programValue->frequency>=bktr->LNB.LNB1 && programValue->frequency<=bktr->LNB.LNB2)  
                return -1;            
            tmpfrequency=( programValue->frequency>bktr->LNB.LNB1 ) ?  (programValue->frequency-bktr->LNB.LNB2):(bktr->LNB.LNB1-programValue->frequency);    
            if ( tmpfrequency <950  || tmpfrequency>2150 )
		        return  -1;
            convertProgram(programValue, &tunerData, &bktr->LNB, bktr->flag22K);
	        tunerData.address = 0xaa;
	        return Lock_Tuner_Program(bktr, programValue, &tunerData);	
        }
        break; 
    case 3:
    case 4:
        {
            if (programValue->frequency>=bktr->LNB.LNB1 && programValue->frequency<=bktr->LNB.LNB2)  
                return -1;                
            tmpfrequency=( programValue->frequency>bktr->LNB.LNB1 ) ?  (programValue->frequency-bktr->LNB.LNB2):(bktr->LNB.LNB1-programValue->frequency);
            if ( tmpfrequency<950  || tmpfrequency>2150 )	
                return  -1;
            new_convertProgram(programValue,(unsigned char *) &new_tunerData, &bktr->LNB, bktr->flag22K);
            return New_Lock_Tuner_Program(bktr, programValue, &new_tunerData);	
        }
        break; 
    case 5:
        {      
            tunerData.frequencyMSB = (u8)((programValue->frequency & 0xff0000) >> 16);
		    tunerData.frequencyLSB = (u8)((programValue->frequency & 0x00ff00) >> 8);
		    tunerData.tunerStep    = (u8)(programValue->frequency & 0x0000ff);
		
		    tunerData.symbolRateHSB = (u8)(programValue->symbolRate >> 8);
		    tunerData.symbolRateMSB = (u8)(programValue->symbolRate & 0x00ff);
		    tunerData.symbolRateLSB = 0;
		    if (programValue->h_V == 256)
    			tunerData.flag = 255;
    		else
    			tunerData.flag = (u8)programValue->h_V;
		    tunerData.address = 0xaa;	
        	tunerSetCheckSum(&tunerData);		    		
            return Lock_Tuner_Program(bktr, programValue, &tunerData);
        }    
        break;      
    case 6:
        {      
            unsigned char *p_new_tunerData = (unsigned char *)&new_tunerData;
            p_new_tunerData[0] = 9;
		    p_new_tunerData[1] = 0;
    		p_new_tunerData[2] = (u8)((programValue->frequency & 0xff0000) >> 16);;
    		p_new_tunerData[3] = (u8)((programValue->frequency & 0x00ff00) >> 8);
    		p_new_tunerData[4] = (u8)(programValue->frequency & 0xff);
    		p_new_tunerData[5] = (u8)((programValue->symbolRate &0xff0000) >> 16);
    		p_new_tunerData[6] = (u8)((programValue->symbolRate &0x00ff00) >> 8);
    		p_new_tunerData[7] = (u8)(programValue->symbolRate & 0xff);
    		if (programValue->h_V == 256)
    			p_new_tunerData[8] = 255;
    		else
    			p_new_tunerData[8] = (u8)programValue->h_V;
            int i;
            int chksum = 0;
            for (i=0; i < 9; i++)  
                chksum += p_new_tunerData[i];
            chksum = ~chksum+1;
            p_new_tunerData[9] = chksum;
            
            return New_Lock_Tuner_Program(bktr, programValue, &new_tunerData);
        }
        break;          
    case 11:
        {
            tunerData.frequencyMSB = (u8)((programValue->frequency & 0xff0000) >> 16);
            tunerData.frequencyLSB = (u8)((programValue->frequency & 0x00ff00) >> 8);
            tunerData.tunerStep = (u8)(programValue->frequency & 0x0000ff);
            tunerData.symbolRateHSB = 0;
            tunerData.symbolRateMSB = (u8)(programValue->symbolRate);		//BindWidth : 6,7,8
        	tunerData.symbolRateLSB = 0;
        	tunerData.flag = 0;        	
        	tunerData.address = 0xaa;	
        	tunerSetCheckSum(&tunerData);
        	return Lock_Tuner_Program(bktr, programValue, &tunerData);	
	    }
	    break;
	}    
	
	return 0;	
}	

static int Lock_Tuner_Program(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue, P_TUNER_DATA tunerData)	
{
    DBGMSG(">>Lock_Tuner_Program()\n"); 
    /*unsigned char *pTest = (unsigned char *)tunerData;
    int k;
    for (k = 0; k < 9; k++)
        TRYMSG("tunerData[%d]=0x%02x \n", k, *pTest++);*/
        
    TUNER_DATA tunerLockData;
    tunerLockData.address=0xab;    
    unsigned char verify = 0;
    int loops;
    int timeout;
    unsigned char *p1, *p2;
    int i;
    int finish;
	    
    loops = 0;
	verify = 0;
    while (verify != 0xff && loops < RETRIES_COUNT)	{
        loops ++;
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*10/100);
        if (i2cWriteMCU(bktr, tunerData ) ){
            verify = 0;
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
        i2cReadVerify( bktr, &verify);
        if (verify != 0xff)  {
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        break;
	}
	releaseGPIO(bktr);
    if (verify != 0xff)    
        return -1;
    finish = 0;
    timeout = 0;
	while (! isGPIOlow(bktr) && timeout++< 800 ){
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
    }
    if (timeout >= 800)        
        return 1;
    timeout = 0;
    while (timeout++ < RETRIES_COUNT ) {
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*3/100);
        if (i2cReadTuner( bktr, &tunerLockData ) )   
            continue;
        releaseGPIO(bktr);
        if (tunerDataCheckSumOK(&tunerLockData) ) {
            finish = 1;
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
    }
    if ( tunerLockData.frequencyLSB+tunerLockData.frequencyMSB ==0 )  
        return -1;
    if (finish){
    	bktr->tflags |= (TUNER_INITALIZED | TUNER_OPEN);
    	p1 = (unsigned char *)programValue;
    	p2 = (unsigned char *)(& bktr->programValue);
    	for (i=0; i<sizeof(PROGRAM_VALUE); i++) 	
    	    *p2++ = *p1++;
        convertTunerData(&tunerLockData, &(bktr->programValue), &(bktr->LNB), programValue->frequency );
        return SUCCESS;
	}
	bktr->tflags &= ~(TUNER_INITALIZED | TUNER_OPEN);
	
	DBGMSG("<<Lock_Tuner_Program()\n"); 
	return -1;
}    

static int New_Lock_Tuner_Program(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue, P_NEW_TUNER_DATA tunerData)
{
    DBGMSG(">>New_Lock_Tuner_Program()\n");
    /*unsigned char *pTest = (unsigned char *)tunerData;
    int k;
    for (k = 0; k < 10; k++)
        TRYMSG("tunerData[%d]=0x%02x \n", k, *pTest++);*/
        
    NEW_TUNER_DATA tunerLockData;
    unsigned char verify = 0;
    int loops;
    int timeout;
    unsigned char *p1, *p2;
    int i;
    int finish;
    
    loops = 0;       
    verify = 0;
    while (verify != 0xff && loops < RETRIES_COUNT) {
        loops ++;
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*10/100);
        interrupt8820(bktr);
        if (new_i2cWriteMCU(bktr, (unsigned char  *) tunerData ) ) {
            reset8820(bktr);
            verify = 0;
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
        new_i2cReadVerify( bktr, &verify);
        if (verify != 0xff) {
            reset8820(bktr);
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        break;
	}
	releaseGPIO(bktr);
    if (verify != 0xff) return -1;
    finish = 0;
    timeout = 0;
    while (! isGPIOlow(bktr) && timeout++< 800 ) {
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
    }
    if (timeout >= 800)    return 1;
    timeout = 0;
    while (timeout++ < RETRIES_COUNT) {
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*3/100);
        if (new_i2cReadTuner( bktr, (unsigned char *) &ca_input) )  
            continue;
        releaseGPIO(bktr);
        if (new_tunerDataCheckSumOK((P_NEW_TUNER_DATA)ca_input) ) {
            strncpy( (char*)(&tunerLockData), ca_input+2, 8);
            finish = 1;
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
 	}
    if ( tunerLockData.frequencyLSB+tunerLockData.frequencyMSB ==0 )  
    {
        return -1;
    }    
    if (finish){
	    bktr->tflags |= (TUNER_INITALIZED | TUNER_OPEN);
	    p1 = (unsigned char *)programValue;
	    p2 = (unsigned char *)(& bktr->programValue);
	    for (i=0; i<sizeof(PROGRAM_VALUE); i++)	
	        *p2++ = *p1++;
        new_convertTunerData(&tunerLockData, &(bktr->programValue), &(bktr->LNB), programValue->frequency );
	    return SUCCESS;
	}
	bktr->tflags &= ~(TUNER_INITALIZED | TUNER_OPEN);
	
	DBGMSG("<<New_Lock_Tuner_Program()\n");
	return -1;
}

static void new_convertProgram(P_PROGRAM_VALUE programValue, unsigned char * tunerData, P_LNB_LOF pLNB, int Flag22K)
{
    DBGMSG(">>new_convertProgram()\n"); 
    
    tunerData[0] = 9;
    tunerData[1] = 0x00;
    tunerData[8] = 0;
	if (programValue->i_Q)	
	    tunerData[8]  |= 0x80;
	if (programValue->symbolRate>8000)	
	    tunerData[8] |= 0x20;
    if ( programValue->h_V)  
        tunerData[8] |= 0x40;
    else  
        tunerData[8] &= 0xbf;
    tunerData[8] &= 0xfc;
    if (Flag22K) tunerData[8] |= 0x01;
    {
	    int tunerStep, divRate;
        int i, frequency ,InFrequency, InNumber;
        frequency = programValue->frequency;
        InFrequency=( frequency > pLNB->LNB1 ) ?  (frequency-pLNB->LNB2):(pLNB->LNB1-frequency);
        tunerStep = 1000000;
	    if(InFrequency <1531)	
	        tunerData[8] |= 0x004;
	    else	
	        tunerData[8] &= 0xfb;
    	InNumber=InFrequency*1000/(tunerStep/1000);
    	tunerData[2]=(InNumber&0x00007f00)>>8;
    	tunerData[3]=InNumber& 0x000000ff;
    	divRate=4000000L/tunerStep;
    	for(i=0;i<16;i++)
	        if(frequencyRef[i]==divRate)	
	            break;
	    if(i>15)
	        i=15;
	    tunerData[4] = i;
        tunerData[2] = (InFrequency >> 8) & 0xff;
        tunerData[3] = InFrequency & 0xff;
	}
    tunerData[5] = (programValue->symbolRate >> 16 ) & 0xff;
    tunerData[6] = (programValue->symbolRate >> 8 ) & 0xff;
    tunerData[7] = programValue->symbolRate & 0xff;
    {  
        int i;
        int chksum = 0;
        for (i=0; i < 9; i++)  
            chksum += tunerData[i];
        chksum = ~chksum+1;
        tunerData[9] = chksum;
    }
        
    DBGMSG("<<new_convertProgram()\n"); 
}

static unsigned char new_tunerDataCheckSumOK(P_NEW_TUNER_DATA tunerData)
{
    DBGMSG(">>new_tunerDataCheckSumOK()\n"); 
        
	unsigned char *pByte = (unsigned char *) tunerData;
	unsigned char chksum = 0;
    int i;
    int length;
    length = pByte[0];
	for ( i=0; i <length; i++) 
	    chksum += pByte[i];
	chksum = ~chksum+1;
	
	DBGMSG("<<new_tunerDataCheckSumOK()\n"); 
	return ( chksum == pByte[length]);
}

static int getTunerProgram(bktr_ptr_t bktr, P_PROGRAM_VALUE programValue)
{
    DBGMSG(">>getTunerProgram()\n");
     
    unsigned char *p1, *p2;
	int i;
	if (bktr->tflags & TUNER_INITALIZED ){
   		p1 = (unsigned char *)programValue;
		p2 = (unsigned char *)(& bktr->programValue);
		for (i=0; i<sizeof(PROGRAM_VALUE); i++)	*p1++ = *p2++;
		    return  SUCCESS;
	}
	
	DBGMSG("<<getTunerProgram()\n");
    return -1;
}	

void dst_close(bktr_ptr_t bktr )
{
    DBGMSG(">>dst_close()\n");
    stop_tuner(bktr);
    DBGMSG("<<dst_close()\n");
}

void bktr_intr( void *arg )
{ 
    //DBGMSG(">>bktr_intr()\n");
    
	bktr_ptr_t		bktr;
	bt848_ptr_t		bt848;
	bt848f1_ptr_t		bt848f1;
	bktr = (bktr_ptr_t) arg;
	bt848 = bktr->base;
	if (bt848->int_mask == ALL_INTS_DISABLED) {
	    return;
	}
    if (bt848->int_stat & BT848_INT_I2CDONE ){
        I2cDone = 1;
        bt848->int_stat =  BT848_INT_I2CDONE;
        bt848->int_mask &= ~ BT848_INT_I2CDONE;
        return;
    }
    bt848f1 = bktr->f1base;
    if (bt848f1->int_stat & (SCERR | OCERR | FDSR | FTRGT | FBUS | RISCI) ) {
        bt848f1->int_stat = (SCERR | OCERR | FDSR | FTRGT | FBUS | RISCI);
    }
    
    return;
}

static int get_ioctl_bytes(void *val, caddr_t arg, int size)
{
        DBGMSG(">>get_ioctl_bytes()\n");
        
#ifdef __FreeBSD__
  memcpy((char *)val, arg, size);
#endif
#ifdef linux2036
  int ret = verify_area(VERIFY_READ, (int *)arg, size);
  if (ret)    return ret;
  memcpy_fromfs(val, (void *)arg, size);
#endif
#ifdef linux
  copy_from_user(val,(void *)arg, size);
#endif

        DBGMSG("<<get_ioctl_bytes()\n");
  return 0;
}

static int put_ioctl_bytes(caddr_t arg, void *val, int size)
{
        DBGMSG(">>put_ioctl_bytes()\n");
        
#ifdef __FreeBSD__
  memcpy(arg, (char *)val, size);
#endif
#ifdef linux2036
  int ret = verify_area(VERIFY_WRITE, (int *)arg, size);
  if (ret)    return ret;
  memcpy_tofs((void *)arg, val, size);
#endif
#ifdef linux
  copy_to_user((void *)arg, val, size);
#endif

        DBGMSG("<<put_ioctl_bytes()\n");
  return 0;
}

#define RISC_INSTRUCTIONS  (AFP_LEN+2)
void get_dmaBufferInfo(bktr_ptr_t bktr, P_DMA_BUFFER_INFO dmaBufferInfo)
{
    DBGMSG(">>get_dmaBufferInfo()\n");
    
    bt848f1_ptr_t		bt848f1;
    u_long              offset;
    int blocks;
    bt848f1 =	bktr->f1base;
    dmaBufferInfo->blocks = DMA_BUFFER_BLOCK;
    dmaBufferInfo->rings = bktr->rings;
    offset =  (unsigned char *)(bt848f1->risc_count) - (unsigned char *)(vtophys(bktr->dma_prog));
    offset /= 8;
    if (offset > 0)
        offset --;
    else
        offset = RISC_INSTRUCTIONS * DMA_BUFFER_BLOCK;
        
    if (offset > 0)   
        offset --;
    else   
        offset = RISC_INSTRUCTIONS * DMA_BUFFER_BLOCK;
        
    blocks = offset / RISC_INSTRUCTIONS;
    offset = offset % RISC_INSTRUCTIONS;
    
    if (offset > AFP_LEN) {
        offset = 0;
        blocks ++;
    }
    
    if (blocks == DMA_BUFFER_BLOCK)  
        blocks = 0;
        
    dmaBufferInfo->currentWrite = blocks * AFP_LEN * (long)ALP_LEN + offset * ALP_LEN;
    dmaBufferInfo->int_stat = bt848f1->int_stat;
    dmaBufferInfo->int_mask = bt848f1->int_mask;
    
    DBGMSG("<<get_dmaBufferInfo()\n");
}

static int dstPutCommand(bktr_ptr_t bktr, P_PROTOCOL protocol)
{
    DBGMSG(">>dstPutCommand()\n");    
    TUNER_DATA tunerData, tunerLockData;
    unsigned char verify = 0;
    int loops;
    int timeout;
    unsigned char *p1, *p2;
    int i;
    int finish = 0;
    p1 = (unsigned char *)(&tunerData);
    p2 = (unsigned char *) protocol;
    for (i=0; i<sizeof(PROTOCOL); i++) 
        p1[i+1] = p2[i];
    tunerSetCheckSum(&tunerData);
    tunerData.address = 0xaa;
    tunerLockData.address=0xab;
    loops = 0;
    verify = 0;
    while (verify != 0xff && loops < RETRIES_COUNT)	{
        loops ++;
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
        interrupt8820(bktr);
        if (i2cWriteMCU(bktr, &tunerData ) )
        {
            reset8820(bktr);
            verify = 0;
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
        i2cReadVerify( bktr, &verify);
        if (verify != 0xff)   {
            reset8820(bktr);
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*5/100);
            continue;
        }
        break;
	}
    if (p2[0]==0 && (p2[1]==1 || p2[1]==3) )  
        return SUCCESS;
    releaseGPIO(bktr);
    if (verify != 0xff)   
        return -1;
    timeout = 0;
	while (! isGPIOlow(bktr) && timeout++< 800 )  {
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
    }
    if (timeout >= 800)   return 1;
    timeout = 0;
    while (timeout++ < RETRIES_COUNT)  {
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
	    if ( i2cReadTuner( bktr, &tunerLockData ) )    
	        continue;
  	    releaseGPIO(bktr);
        if (tunerDataCheckSumOK(&tunerLockData) ) {
            finish = 1;
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
 	}
	if (finish){
        p1 = (unsigned char *)(&tunerLockData);
        p1 ++;
        p2 =  bktr->commandReply;
        for (i=0; i<sizeof(PROTOCOL); i++)	
            *p2++ = *p1++;
        return SUCCESS;         
	}
	DBGMSG("<<dstPutCommand()\n");
	return -1;
}

static int dstCommandReply(bktr_ptr_t bktr, P_PROTOCOL protocol)
{
    DBGMSG(">>dstCommandReply()\n");
        
	unsigned char *p1, *p2;
	int i;
    p1 = (unsigned char *)protocol;
	p2 =  bktr->commandReply;
	for (i=0; i<sizeof(PROTOCOL); i++) 
	    *p1++ = *p2++;
	
	DBGMSG("<<dstCommandReply()\n");
	return  SUCCESS;
}	

void makenullEvent(void)
{
    DBGMSG(">>makenullEvent()\n");
    
    ca_input[0] = 7;
    ca_input[1] = 0x40;
    ca_input[2] = ca_input[3] = 0;
    ca_input[4] = 0;
    ca_input[5] = ca_input[6] = 0;
    ca_input[7] = 0xb9;
    
    DBGMSG("<<makenullEvent()\n");
}

static void interrupt8820(bktr_ptr_t bktr)
{
    DBGMSG(">>interrupt8820()\n");
    
    int i;
    release8820(bktr);
    for (i=0;i<50;i++);
    inter8820(bktr);
    
    DBGMSG("<<interrupt8820()\n");
}

#define PCMSG_NULL                 0x00
#define PCMSG_APPLICATION_INFO     0x01
#define PCMSG_CA_INFO              0x02
#define PCMSG_CA_PMT               0x03
#define PCMSG_CA_PMT_REPLY         0x04
#define PCMSG_DATETIME_ENQ         0x05
#define PCMSG_DATETIME             0x06
#define PCMSG_ENQ                  0x07
#define PCMSG_ANSWER               0x08
#define PCMSG_ENTER_MENU           0x09
#define PCMSG_MENU           	   0x0A
#define PCMSG_MENU_ANSWER          0x0B
#define PCMSG_LIST                 0x0C
#define PCMSG_GET_MMI              0x0D
#define PCMSG_CLOSE_MMI		   0x0e

static int put8820CiMSG(bktr_ptr_t bktr, unsigned char *command)
{
    DBGMSG(">>put8820CiMSG()\n");
        
    unsigned char verify = 0;
    int loops;
    int timeout;
    //unsigned char *p1, *p2;
    //int i;
    int finish = 0;
    loops = 0;
    verify = 0;
    while (verify != 0xff && loops < RETRIES_COUNT)
    {
        loops ++;
        pulldownGPIO(bktr);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*10/100);
        interrupt8820(bktr);
        if (new_i2cWriteMCU(bktr, (unsigned char  *) command ) ){
            reset8820(bktr);
            verify = 0;
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
        new_i2cReadVerify( bktr, &verify);
        if (verify != 0xff)
        {
            reset8820(bktr);
            releaseGPIO(bktr);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100);
            continue;
        }
        break;
    }
	releaseGPIO(bktr);
    if (verify != 0xff) 
        return -1;
    switch (command[4])
    {
        case PCMSG_APPLICATION_INFO:
        case PCMSG_CA_INFO:
        case PCMSG_CA_PMT_REPLY:
        case PCMSG_DATETIME_ENQ:
        case PCMSG_GET_MMI:
              break;
        default:
             return SUCCESS;
    }
    finish = 0;
    timeout = 0;
    while (! isGPIOlow(bktr) && timeout++< 8000 ) {
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*1/100);
    }
    if (timeout >= 8000) return 1;
    timeout = 0;
    while (timeout++ < RETRIES_COUNT){
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*3/100);      
        if (new_i2cReadTuner( bktr, (unsigned char *) ca_input ) )  
            continue;
        if (new_tunerDataCheckSumOK((P_NEW_TUNER_DATA)ca_input) ) {
            finish = 1;
            break;
        }
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ*5/100);
    }
    if (finish) {
        ca_status = HAS_MSG;
		return SUCCESS;
	}
    ca_status = ERROR_STATUS;
        
    DBGMSG("<<put8820CiMSG()\n");
	return -1;
}

static void reset8820(bktr_ptr_t bktr)
{
    DBGMSG(">>reset8820()\n");
    
    bt848_ptr_t	bt848 = bktr->base;
    //int i;
    bt848->gpio_out_en |= 0x00000004;
    bt848->gpio_data &= ~0x04;
    udelay(10000);
    bt848->gpio_data |= 0x04;
    udelay(10000);
    bt848->gpio_out_en &= ~0x000004;
    
    DBGMSG("<<reset8820()\n");
}

int dst_ioctl( bktr_ptr_t bktr, int unit, int cmd, caddr_t arg, struct proc* pr )
{
    DBGMSG(">>dst_ioctl()\n");
    
    int	ret = 0;
    bt848_ptr_t      bt848;
    LNB_LOF         tmpLNB;
    PROTOCOL        protocol;
    PROGRAM_VALUE	programValue;
    DMA_BUFFER_INFO dmaBufferInfo;
    DEMULTIPLEXER_VALUE      DMV;
    DISEQC_VALUE  DV;
    SIGNAL_VALUE  SV;
    CARD_VALUE    CV;
    LNB_INFO      LnbInfo;
    DISEQC_COMMAND DiscComm;
    
    //int timeout;
	bt848 =	bktr->base;
	switch ( cmd ) {
        case TUNER_SETTUNER:
            if ( (ret = get_ioctl_bytes(&programValue, arg, sizeof(programValue) ) )) 
                return ret;
            
            return Lock_Tuner(bktr, &programValue);
                 
            break;
		case TUNER_GETTUNER:
			ret = getTunerProgram(bktr, &programValue);
            if ( (ret = put_ioctl_bytes(arg, &programValue, sizeof(programValue) ) )) 
                return ret;
            break;
		case TUNER_STOP:
			return stop_tuner(bktr);
		case DST_INIT:
		    if ((bktr->card_type != 0) && (bktr->card_type != 1) && (bktr->card_type != 2) && (bktr->card_type != 3) && (bktr->card_type != 4))
		        return -1;
            if ( (ret = get_ioctl_bytes(&tmpLNB, arg, sizeof(LNB_LOF) ) ))	
                return ret;
            if (tmpLNB.LNB1 < tmpLNB.LNB2){
                bktr->LNB.LNB1 = tmpLNB.LNB1;  bktr->LNB.LNB2 = tmpLNB.LNB2;
            } else {
                bktr->LNB.LNB1 = tmpLNB.LNB2;   
                bktr->LNB.LNB2 = tmpLNB.LNB1;
            }
            break;
		case DST_GET_STATUS:
			break;
		case DST_START_CAPTURE:
			start_capture(bktr); 
            break;
		case DST_STOP_CAPTURE:
		    //stop_capture(bktr); //louis: If enable it, driver will crash..
            break;
		case DST_GET_BUFFER_INFO:
            get_dmaBufferInfo(bktr, &dmaBufferInfo);
            if ( (ret = put_ioctl_bytes(arg, &dmaBufferInfo, sizeof(dmaBufferInfo) ) )) 
                return ret;
            break;
        case DST_PUT_COMMAND:
            if ( (ret = get_ioctl_bytes(&protocol, arg, sizeof(PROTOCOL) ) )) 
                return ret;
            return dstPutCommand(bktr, &protocol);
            break;
        case DST_GET_REPLY:
            dstCommandReply(bktr, &protocol);
            if ( (ret = put_ioctl_bytes(arg, protocol, sizeof(PROTOCOL) ) ) ) 
                return ret;
            break;
        case WRITE_CAMSG:
            if ( (ret = get_ioctl_bytes(&ca_output, arg, sizeof(STRING) ) )) 
                return ret;
            return put8820CiMSG(bktr, ca_output);
			break;
        case READ_CAMSG:
            if (ca_status == HAS_MSG) {
                if ( (ret = put_ioctl_bytes(arg, ca_input+2, sizeof(STRING) ) ) ) 
                    return ret;
            }        
            break;
        case RESET_MCU:
            reset8820(bktr);
            break;

		case NET_ADD_IF:
		{
            dst_net_if  dstnetif;
            int result;

            if(copy_from_user(&dstnetif,(struct dst_net_if*)arg, sizeof(dst_net_if)))
                return -EFAULT;
            result=dst_net_add_if(dstnetif.pid);
            if (result<0)
                return result;
            dstnetif.if_num=result;
            // if(copy_to_user((struct dst_net_if*)arg, &dstnetif, sizeof(dst_net_if)))
            //       return -EFAULT;
            break;
        }
        case NET_REMOVE_IF:
		    dst_net_remove_if((int)arg);
		    break;  
		    
		case INIT_DEMUX:
		    if ((ret = get_ioctl_bytes(&DMV, arg, sizeof(DEMULTIPLEXER_VALUE)))) 
                return ret;
		    if ((ret = Init_Demux(bktr, &DMV, unit)))
		        return ret;
		    break;
		  
		case DST_SET_VOLTAGE:
		    if ((bktr->card_type != 0) && (bktr->card_type != 1) && (bktr->card_type != 2) && (bktr->card_type != 3) && (bktr->card_type != 4))
		        return -1;
		    return dst_set_voltage(bktr, (int)arg);
		    break;
		    
		case DST_SET_DISEQC:    
		    if ((bktr->card_type != 0) && (bktr->card_type != 1) && (bktr->card_type != 2) && (bktr->card_type != 3) && (bktr->card_type != 4))
		        return -1;
		    if ( (ret = get_ioctl_bytes(&DV, arg, sizeof(DISEQC_VALUE)))) 
                return ret;
            return dst_set_diseqc(bktr, &DV);
            break;
            
        case DST_GET_SIGNAL:  
            if (dst_get_signal(bktr, &SV))
                return ret;
            if ((ret = put_ioctl_bytes(arg, &SV, sizeof(SIGNAL_VALUE)))) 
                return ret;
            break;
            
         case DST_GET_CARD_TYPE:
             memcpy(CV.sCard_Name, bktr->card_name, 6);
             CV.sCard_Name[6] = '\0';
             CV.nTS_SIZE = bktr->TS_Size;
            if ((ret = put_ioctl_bytes(arg, &CV, sizeof(CARD_VALUE)))) 
                return ret;
            break;
         
         case DST_SET_LNB:
            if ( (ret = get_ioctl_bytes(&LnbInfo, arg, sizeof(LNB_INFO)))) 
                return ret;
            return dst_set_lnb(bktr, &LnbInfo);
            break;         
            
         case DST_SET_DISEQC_COMMAND:
            if ( (ret = get_ioctl_bytes(&DiscComm, arg, sizeof(DISEQC_COMMAND)))) 
                return ret;
            return dst_set_diseqc_command(bktr, &DiscComm);
            break;            
	}       
	
    DBGMSG("<<dst_ioctl()\n");
    return SUCCESS;
}

///////////////////////////////////dst card virtual worknet driver part //////////////////////////////////////
static u_long Getdmabufferpos(void)    // used by IP
{
    DBGMSG(">>Getdmabufferpos()\n");
    
    u_long IP_dmaoffset;
    bktr_ptr_t bktr;
    bt848f1_ptr_t		bt848f1;
    u_long                  offset;
    int blocks;
    
    bktr = &brooktree[0];
    bt848f1 =bktr->f1base;
    
    offset =  (unsigned char *)(bt848f1->risc_count) - (unsigned char *)(vtophys(bktr->dma_prog));
    offset /= 8;                // change to instruction's order  ????

    if (offset > 0)
        offset --;                  // back 1 instruction
    else  // last instruction in capture loop
        offset = RISC_INSTRUCTIONS * DMA_BUFFER_BLOCK;

    if (offset > 0)
        offset --;       // back 1 more instruction for write completion
    else  // last instruction in capture loop
        offset = RISC_INSTRUCTIONS * DMA_BUFFER_BLOCK;

    blocks = offset / RISC_INSTRUCTIONS;
    offset = offset % RISC_INSTRUCTIONS;

    // test position
    if (offset > AFP_LEN)
    {
        offset = 0; //255;
        blocks ++;
    }
    if (blocks == DMA_BUFFER_BLOCK) // current just write to the end of buffer
    {
        blocks = 0;
    }
    IP_dmaoffset=blocks*AFP_LEN*(long)ALP_LEN + offset*ALP_LEN;

    //printk("****IP_dmaoffset=%x\n",  IP_dmaoffset);
    
    DBGMSG("<<Getdmabufferpos()\n");
    return IP_dmaoffset;
}

unsigned short my_eth_type_trans(struct sk_buff *skb, struct net_device *dev)
{
    DBGMSG(">>my_eth_type_trans()\n");
        
	struct ethhdr *eth;
	unsigned char *rawp;

	skb->mac.raw=skb->data;
	skb_pull(skb,dev->hard_header_len);
	eth= skb->mac.ethernet;

	if(*eth->h_dest & 1)
	{
		if(memcmp(eth->h_dest,dev->broadcast, ETH_ALEN)==0)
			skb->pkt_type=PACKET_BROADCAST;
		else
			skb->pkt_type=PACKET_MULTICAST;
	}

	if (ntohs(eth->h_proto) >= 1536)
		return eth->h_proto;

	rawp = skb->data;

	/*
	 *	This is a magic hack to spot IPX packets. Older Novell breaks
	 *	the protocol design and runs IPX over 802.3 without an 802.2 LLC
	 *	layer. We look for FFFF which isn't a used 802.2 SSAP/DSAP. This
	 *	won't work for fault tolerant netware but does for the rest.
	 */
	if (*(unsigned short *)rawp == 0xFFFF)
		return htons(ETH_P_802_3);

	/*
	 *	Real 802.2 LLC
	 */
	 
	DBGMSG("<<my_eth_type_trans()\n");
	return htons(ETH_P_802_2);
}

static void dst_net_sec(struct net_device *dev, u8 *s_pkt, int s_pkt_len)
{
    DBGMSG(">>dst_net_sec()\n");
    
    u8 *eth;
    struct sk_buff *skb;
    u8 *pkt;
    int pkt_len;

    pkt=s_pkt+(*s_pkt+1);
    pkt_len=s_pkt_len-(*s_pkt+1);
    printk("table_id=0x%x  pkt_len=%d  ",pkt[0],pkt_len);
    /*if( (IP_mac[0]!=pkt[0x0b]) ||
        (IP_mac[1]!=pkt[0x0a]) ||
        (IP_mac[2]!=pkt[0x09]) ||
        (IP_mac[3]!=pkt[0x08]) ||
        (IP_mac[4]!=pkt[0x04]) ||
        (IP_mac[5]!=pkt[0x03]) )
        return ;    */
    //printk("mac= %d=   %d=   %d=    %d=   %d=   %d=  \n",
    //       pkt[0x0b], pkt[0x0a],pkt[0x09],pkt[0x08],pkt[0x04], pkt[0x03]);

    printk("section_length=%d\n",((pkt[1] & 0x0f)<<8) | pkt[2]);

	if (!pkt_len)
        return;
    skb = dev_alloc_skb(pkt_len+2);
    if (skb == NULL) {
        printk(KERN_NOTICE "%s: Memory squeeze, dropping packet.\n",dev->name);
        ((dst_net_priv_t *)dev->priv)->stats.rx_dropped++;
        return;
    }
    eth=(u8 *) skb_put(skb,pkt_len+2);
    memcpy(eth+14, (void*)pkt+12, pkt_len-12);

    eth[0]=pkt[0x0b];
    eth[1]=pkt[0x0a];
    eth[2]=pkt[0x09];
    eth[3]=pkt[0x08];
    eth[4]=pkt[0x04];
    eth[5]=pkt[0x03];
    eth[6]=eth[7]=eth[8]=eth[9]=eth[10]=eth[11]=0;
    eth[12]=0x08; eth[13]=0x00;

	skb->protocol=my_eth_type_trans(skb,dev);  //*****
    skb->dev=dev;

    ((dst_net_priv_t *)dev->priv)->stats.rx_packets++;
    ((dst_net_priv_t *)dev->priv)->stats.rx_bytes+=skb->len;
    printk("skb->len=%d\n",skb->len);
    // sti();  //?????
    netif_rx(skb);
    
    DBGMSG("<<dst_net_sec()\n");
}

static inline unsigned char packet_read_char(demuxer_t *demuxer)
{
    DBGMSG(">>packet_read_char()\n");
        
	unsigned char result;
    u_long  pos;

	pos=demuxer->pos+demuxer->offset;
	result=*(demuxer->pbuf+pos);  demuxer->offset++;

    DBGMSG("<<packet_read_char()\n");
	return result;
}

static inline unsigned int packet_read_int16(demuxer_t *demuxer)
{
    DBGMSG(">>packet_read_int16()\n");
        
	unsigned int a, b, result;
    u_long pos;

	pos=demuxer->pos+demuxer->offset;
	a =*(demuxer->pbuf+pos); demuxer->offset++;

    pos=demuxer->pos+demuxer->offset;
	b =*(demuxer->pbuf+pos); demuxer->offset++;

	result = (a << 8) | b;
	
	DBGMSG("<<packet_read_int16()\n");
	return result;
}

static inline unsigned int packet_read_int24(demuxer_t *demuxer)
{
    DBGMSG(">>packet_read_int24()\n");
        
	unsigned int a, b, c, result;
    u_long pos;

	pos=demuxer->pos+demuxer->offset;
	a =*(demuxer->pbuf+pos); demuxer->offset++;

    pos=demuxer->pos+demuxer->offset;
	b =*(demuxer->pbuf+pos); demuxer->offset++;

    pos=demuxer->pos+demuxer->offset;
	c = *(demuxer->pbuf+pos); demuxer->offset++;

	result = (a << 16) | (b << 8) | c;

    DBGMSG("<<packet_read_int24()\n");
	return result;
}

static inline unsigned int packet_read_int32(demuxer_t *demuxer)
{
    DBGMSG(">>packet_read_int32()\n");
        
	unsigned int a, b, c, d, result;
    u_long pos;

	pos=demuxer->pos+demuxer->offset;
	a =*(demuxer->pbuf+pos); demuxer->offset++;

	pos=demuxer->pos+demuxer->offset;
	b =*(demuxer->pbuf+pos); demuxer->offset++;

	pos=demuxer->pos+demuxer->offset;
	c =*(demuxer->pbuf+pos); demuxer->offset++;

	pos=demuxer->pos+demuxer->offset;
	d =*(demuxer->pbuf+pos); demuxer->offset++;

	result = (a << 24) | (b << 16) | (c << 8) | d;
	
	DBGMSG("<<packet_read_int32()\n");
	return result;
}

static unsigned int packet_skip(demuxer_t *demuxer, long length)
{
    DBGMSG(">>packet_skip()\n");
        
	demuxer->offset += length;
	
	DBGMSG("<<packet_skip()\n");
	return 0;
}

static int get_adaptation_field(demuxer_t *demuxer)
{
    DBGMSG(">>get_adaptation_field()\n");
        
	long    length;
	int     pcr_flag;

	length = packet_read_char(demuxer);
	if(length>0) {
        pcr_flag = (packet_read_char(demuxer) >> 4) & 1;
        if(pcr_flag)
        {
            packet_read_int32(demuxer);
            packet_read_int16(demuxer);
            if(length)
                packet_skip(demuxer, length - 7);
        }
        else
            packet_skip(demuxer, length - 1);
    }
        
    DBGMSG("<,get_adaptation_field()\n");
	return 0;
}

static int packet_get_data_buffer(demuxer_t *demuxer)
{
    DBGMSG(">>packet_get_data_buffer()\n");
        
    int len;
    u_long  scr_pos;

	len = 188 - demuxer->offset;
    scr_pos=demuxer->pos+demuxer->offset;

	if(demuxer->pid==IP_pid)
	{
	    memcpy(demuxer->section+demuxer->section_size, demuxer->pbuf+scr_pos, len);
        demuxer->section_size += len;
    }
	demuxer->offset += len;
	
	DBGMSG("<<packet_get_data_buffer()\n");
	return 0;
}

static int get_payload(demuxer_t *demuxer)
{
    DBGMSG(">>get_payload()\n");
    
    if(demuxer->pid == IP_pid)
        packet_get_data_buffer(demuxer);
    else
        packet_skip(demuxer, TS_SIZE - demuxer->offset);
    
    DBGMSG("<<get_payload()\n");   
    return 0;
}

static int TS_packet_parse(struct net_device *dev,u_long pos, uint8_t *pbuf)
{
    DBGMSG(">>TS_packet_parse()\n");
    
    unsigned int bits;
    
    demuxer->pbuf = pbuf;
    demuxer->pos = pos;
    demuxer->offset = 0;
    
    if(packet_read_char(demuxer) != 0x47)
        return 1;
    bits = packet_read_int24(demuxer) & 0x00ffffff;
    demuxer->pid = (bits >> 8) & 0x00001fff;
    if(demuxer->pid !=IP_pid) 
        return 0;

	demuxer->transport_error_indicator = (bits >> 23) & 0x1;
    demuxer->payload_unit_start_indicator = (bits >> 22) & 0x1;
    demuxer->transport_scrambling_control = (bits >> 6) & 0x3;
    demuxer->adaptation_field_control = (bits >> 4) & 0x3;
    demuxer->continuity_counter = bits & 0xf;

	if(demuxer->payload_unit_start_indicator)
	{
        if(demuxer->section_valid==1)
	    {
            printk(" demuxer->section_size=%d  \n",demuxer->section_size);
            dst_net_sec(dev,demuxer->section, demuxer->section_size); //****
	    }
        demuxer->section_valid=1;
	    demuxer->section_size=0;
	}
	
    if(demuxer->adaptation_field_control == 2 || demuxer->adaptation_field_control == 3)
        get_adaptation_field(demuxer);
    if(demuxer->adaptation_field_control == 1 || demuxer->adaptation_field_control == 3)
	    get_payload(demuxer);
    	
    DBGMSG("<<TS_packet_parse()\n");    
	return  0;
}

static u_long demux(struct net_device *dev, u_long startpos, u_long endpos)
{
    DBGMSG(">>demux()\n");
        
    u_long pos;
    
    pos = startpos;
    if (startpos < endpos)
    {
        //printk("startpos=%d    endpos=%d\n ", startpos , endpos);
        while ((endpos >= (pos + TS_SIZE)) ? 1:0)
        {
        if (*(mmbuf + pos)==0x47)
        {
            if(TS_packet_parse(dev, pos, mmbuf)==1)
                pos++;
	        else
	            pos += TS_SIZE;
        }
        else
            pos++;
        }
        if (pos>=DMABUFFERLEN)
        pos=0;
    }
    else if (startpos>endpos)
    {
        while ((DMABUFFERLEN >= (pos + TS_SIZE))? 1:0)
        {
            if (*(mmbuf + pos) == 0x47)
            {
                if(TS_packet_parse(dev, pos, mmbuf) == 1)
                    pos++;
    	        else
    	            pos += TS_SIZE;
	        }
	        else 
	        {
                pos++;
	        }
	    }
    	if (pos >= DMABUFFERLEN)   
    	    pos = 0;
    	if (endpos < TS_SIZE-1) 
    	    return pos;
    	if (pos != 0)
	    {
            memmove(tsbuf, mmbuf + pos, DMABUFFERLEN - pos);
            memmove(tsbuf + DMABUFFERLEN - pos, mmbuf, TS_SIZE - (DMABUFFERLEN - pos));

	        if (TS_packet_parse(dev, 0, tsbuf) == 1) ; //Louis: ??
            pos = TS_SIZE - (DMABUFFERLEN - pos);
        }
        while ((endpos >= (pos + TS_SIZE - 1))? 1:0)
        {
            if (*(mmbuf+pos) == 0x47)
	        {
                if (TS_packet_parse(dev, pos, mmbuf) == 1)
                    pos++;
	            else
	                pos += TS_SIZE;
	        }
	        else
	        { 
	            pos++;
	        }
        }

	    if (pos >= DMABUFFERLEN)
            pos=0;
    }
   
   DBGMSG("<<demux()\n");
   return pos;
}

static int section_thread( void *netdev )
{
    DBGMSG(">>section_thread()\n");
      
    u_long  pre_pos,cur_pos;
    
    demuxer=(demuxer_t *)vmalloc(sizeof(demuxer_t));
    demuxer->section_size=0;
    demuxer->section_valid=0;
    
    IP_free_flag=0;
    pre_pos=Getdmabufferpos();
    while(1)
    {
        ddelay(100);
	    if( IP_free_flag==1)
	        break;

	    cur_pos= Getdmabufferpos();
        if(cur_pos==pre_pos)
	    {
            ddelay(100);
	        continue;
	    }
        pre_pos=demux((struct net_device *)netdev,pre_pos,cur_pos);
    }
      
    DBGMSG("<<section_thread()\n");
    return 0;
}

static int dst_net_filter_set(struct net_device *dev, unsigned char *mac)
{
    DBGMSG(">>dst_net_filter_set()\n");
     
    dst_net_priv_t *priv=(dst_net_priv_t *)dev->priv;

    if (!(mac[0]|mac[1]|mac[2]|mac[3]|mac[4]|mac[5]))
        return 0;

    IP_pid=priv->pid;
    IP_mac[0]=mac[0];
    IP_mac[1]=mac[1];
    IP_mac[2]=mac[2];
    IP_mac[3]=mac[3];
    IP_mac[4]=mac[4];
    IP_mac[5]=mac[5];
    
    kernel_thread(section_thread,(struct net_device *)dev,0);
    
    DBGMSG("<<dst_net_filter_set()\n");
    return  0;
}

static void dst_net_filter_free(struct net_device *dev)
{
    DBGMSG(">>dst_net_filter_free()\n");
    
    IP_free_flag=1;
    ddelay(1000);  //1000 ms
    kfree(demuxer);
    
    DBGMSG("<<dst_net_filter_free()\n");
}

static int dst_net_tx(struct sk_buff *skb, struct net_device *dev)
{
    DBGMSG(">>dst_net_tx()\n");
    DBGMSG("<<dst_net_tx()\n");
	return 0;
}

static void dst_set_mc_filter(struct net_device *dev, struct dev_mc_list *mc)
{
    DBGMSG(">>dst_set_mc_filter()\n");
        
	printk("%s: set_mc_filter %02x %02x %02x %02x %02x %02x\n",
	dev->name,
	mc->dmi_addr[0],
	mc->dmi_addr[1],
	mc->dmi_addr[2],
	mc->dmi_addr[3],
	mc->dmi_addr[4],
	mc->dmi_addr[5]);
	
	DBGMSG("<<dst_set_mc_filter()\n");
}

static void  dst_net_set_multi(struct net_device *dev)
{
    DBGMSG(">>dst_net_set_multi()\n");
        
	printk("%s: set_multi\n", dev->name);

	if (dev->flags & IFF_PROMISC)
	{
    	//printk("%s: promiscuous mode\n", dev->name);
    	//priv->mode=3;
	}
	else if((dev->flags &IFF_ALLMULTI))
	{
		//printk("%s: allmulti mode\n", dev->name);
		//priv->mode=2;
	}
	else if(dev->mc_count)
	{
        int mci=0;
        struct dev_mc_list *mc;

		for (mc=dev->mc_list; (mc!=NULL) &&  (mci<DST_NET_MULTICAST_MAX);mc=mc->next, mci++)
		{
		    dst_set_mc_filter(dev, mc);
        }
	}
	
	DBGMSG("<<dst_net_set_multi()\n");
}

static int dst_net_set_config(struct net_device *dev, struct ifmap *map)
{
    DBGMSG(">>dst_net_set_config()\n");
        
    printk(" ****** dst_net_set_config()");
	if (netif_running(dev))
	    return -EBUSY;
	//printk("dvb_net: PID=%04x\n", map->base_addr);
	
	DBGMSG("<<dst_net_set_config()\n");
	return 0;
}

static int dst_net_set_mac(struct net_device *dev, void *p)  // " p " come from ?
{
    DBGMSG(">>dst_net_set_mac()\n");
        
	struct sockaddr *addr=p;
    unsigned char *mac;

	printk("%s: set_mac\n", dev->name);
	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
	mac=(unsigned char *) dev->dev_addr;
	if (netif_running(dev))
	{
	    dst_net_filter_free(dev);
	    if(dst_net_filter_set(dev,mac))  //????
	        printk("dst_net_filter_set failed\n");
	}
	
	DBGMSG("<<dst_net_set_mac()\n");
    return 0;
}

static int dst_net_open(struct net_device *dev)
{
    DBGMSG(">>dst_net_open()\n");
        
    printk("**** dst_net_open() \n");
	if(dst_net_filter_set(dev,dev->dev_addr))
	    printk("dst_net_filter_set failed\n");
	   
	DBGMSG("<<dst_net_open()\n");   
	return 0;
}

static int dst_net_stop(struct net_device *dev)
{
    DBGMSG(">>dst_net_stop()\n");
        
    dst_net_filter_free(dev);
	printk("dst_net: stop\n");
	
	DBGMSG("<<dst_net_stop()\n");
	return 0;
}
static struct net_device_stats *dst_net_get_stats(struct net_device *dev)
{
    DBGMSG(">>dst_net_get_stats()\n");
    DBGMSG("<<dst_net_get_stats()\n");
    return &((dst_net_priv_t *)dev->priv)->stats;
}

static int dst_net_init_dev(struct net_device *dev)
{
        DBGMSG(">>dst_net_init_dev()\n");
        printk("****** dst_net_init_dev()\n");

	dev->open		= dst_net_open;
	dev->stop		= dst_net_stop;
	dev->hard_start_xmit	= dst_net_tx;
	dev->hard_header	= 0;
	dev->get_stats		= dst_net_get_stats;
	dev->set_multicast_list = dst_net_set_multi; //when to exec this function ?
	dev->set_config         = dst_net_set_config;
        ether_setup(dev);
	dev->set_mac_address    = dst_net_set_mac;  //when to exec this function ?
	dev->mtu		= 4096;

	//dev->mc_count=0;
	//dev->flags             |= IFF_NOARP;
	//dev->hard_header_cache  = NULL;

        DBGMSG("<<dst_net_init_dev()\n");
	return 0;
}

static int get_if(void)
{
    DBGMSG(">>get_if()\n");
        
	int i;

	for (i=0; i<dstnet->dev_num; i++)
	  if (!dstnet->state[i])  
	    break;
	if (i==dstnet->dev_num)  
	    return -1;
	dstnet->state[i]=1;
	
	DBGMSG("<<get_if()\n");
	return i;
}

static int dst_net_add_if(u16 pid)   //ioctl function
{
    DBGMSG(">>dst_net_add_if()\n");
        
    struct net_device *net;
	int result;
	int if_num;

	if_num=get_if();
	if (if_num<0)  
	    return -EINVAL;

	net=&dstnet->device[if_num];
	//demux=dvbnet->demux;

	net->base_addr = 0;
	net->irq       = 0;
	net->dma       = 0;
	net->mem_start = 0;
    memcpy(net->name, "dst0_0", 7);
    net->name[3]=dstnet->card_num+0x30;
    net->name[5]=if_num+0x30;
    //printk("****** net->name=%s\n",net->name);
	net->next      = NULL;
    net->init      = dst_net_init_dev; //******
	net->priv      = kmalloc(sizeof(dst_net_priv_t), GFP_KERNEL);  //???
	if (net->priv == NULL)
		return -ENOMEM;

	memset(net->priv, 0, sizeof(dst_net_priv_t));

    ((dst_net_priv_t *)net->priv)->pid=pid;

	net->base_addr=pid;  //?????

	if ((result = register_netdev(net)) < 0)
	    return result;
	MOD_INC_USE_COUNT;
//printk("***  if_num=%d   pid=0x%x\n",if_num,net->base_addr);

    DBGMSG("<<dst_net_add_if()\n");
	return if_num;
}

static int dst_net_remove_if(int num)   //ioctl call this function
{
    DBGMSG(">>dst_net_remove_if()\n");
        
	if (!dstnet->state[num])
		return -EINVAL;
	dst_net_stop(&dstnet->device[num]);
    kfree(dstnet->device[num].priv);
    unregister_netdev(&dstnet->device[num]);
	dstnet->state[num]=0;
	MOD_DEC_USE_COUNT;
	
	DBGMSG("<<dst_net_remove_if()\n");
	return 0;
}

void dst_net_release(void)   
{
    DBGMSG(">>dst_net_release()\n");
        
	int i;

	for (i=0; i<dstnet->dev_num; i++)
	{
        if (!dstnet->state[i])
		    continue;
		dst_net_remove_if(i);
	}
	vfree(dstnet);  //kfree(dstnet);
    printk("****IP:dst_net_release()\n");
        
    DBGMSG("<<dst_net_release()\n");
}

    //0:        : Old DST Card				--- 188bytes;
	//1: ST-MOT : New DST Card(USALS)			--- 188bytes;
	//2: ST-03T : New DST Card(USALS, 3 Tuners)--- 204bytes;
	//3: DST-CI : DST CI Card					--- 204bytes;
	//4: DSTMCI : DST CI Card(188 Bytes)		--- 188bytes;
	//5: DCTNEW : DCT Card						--- 188bytes;
	//6: DCT-CI : DCT CI Card					--- 204bytes;
	//10:DTTNXT : DTT Card(Digtal & Analog)
	//11:DTTDIG : DTT Card(Digtal)
int  check_card_version(bktr_ptr_t bktr)
{
     DBGMSG(">>check_card_version()\n");
        
     PROTOCOL comm={0x00,0x06,0x00,0x00,0x00,0x00,0x00,0xfa} ;
     if ( dstPutCommand(bktr, &comm) != 0 ){
        bktr->card_version = 0;
        return -1;
     }     
     dstCommandReply(bktr, &comm);
     
     if(strncmp( comm+1, "ST-MOT", 6) == 0){
        bktr->card_version = 0;
        bktr->TS_Size = 188;
        bktr->card_type = 1;
     } else if (strncmp( comm+1, "ST-03T", 6) == 0){
        bktr->card_version = 0;
        bktr->TS_Size = 204;     
        bktr->card_type = 2;
     } else if ( strncmp( comm+1, "DST-CI", 6) == 0){        
        bktr->card_version = 1;
        bktr->TS_Size = 204;        
        bktr->card_type = 3;
     } else  if ( strncmp( comm+1, "DSTMCI", 6) == 0){
        bktr->card_version = 1;
        bktr->TS_Size = 188;        
        bktr->card_type = 4;
     } else if (strncmp( comm+1, "DCTNEW", 6) == 0){
        bktr->card_version = 0;
        bktr->TS_Size = 188;	
        bktr->card_type = 5;
     } else if (strncmp( comm+1, "DCT-CI", 6) == 0){
        bktr->card_version = 1;
        bktr->TS_Size = 204;	
        bktr->card_type = 6;
     } else if (strncmp( comm+1, "DTTNXT", 6) == 0){
        bktr->card_version = 0;
        bktr->TS_Size = 188;	     
        bktr->card_type = 10;
     } else if (strncmp( comm+1, "DTTDIG", 6) == 0){
        bktr->card_version = 0;
        bktr->TS_Size = 188;
        bktr->card_type = 11;	    
     }
          
     memcpy(bktr->card_name, comm+1, 6);
     TS_SIZE = bktr->TS_Size;
     
     DBGMSG("<<check_card_version()\n");
     return bktr->card_version;
}
int dst_net_init(void)
{
    DBGMSG(">>dst_net_init()\n");
        
	int i;

	dstnet=(dst_net_t *)vmalloc(sizeof(dst_net_t));
	dstnet->card_num=0;       //***
	dstnet->dev_num=DST_NET_DEVICES_MAX;  //DST_NET_DEVICES_MAX=10
	for (i=0; i<dstnet->dev_num; i++)
	    dstnet->state[i]=0;  //not be used
	printk("**** dst_net_init()\n");

    DBGMSG("<<dst_net_init()\n");
	return 0;
}

int dst_set_voltage(bktr_ptr_t bktr, int nVol)
{
    DBGMSG(">>dst_set_voltage() nVol=%d\n", nVol);    
    
    PROTOCOL power_off = {0x0, 0x09, 0xff, 0xff, 0x00, 0x00, 0x00, 0x0}; //X3=00 power off
    PROTOCOL power_on = {0x0, 0x09, 0xff, 0xff, 0x01, 0x00, 0x00, 0x0}; //X3=01 power on
    
    switch(nVol) {
    case 1:
        if (dstPutCommand(bktr, &power_on) != 0)
            return -1;
        break;
    case 0:
        if (dstPutCommand(bktr, &power_off) != 0)
            return -1;
        break;
    default:    
        DBGMSG("dst_set_voltage => Set error !(n=1~2)\n");        
        return -1;
    }
    
    DBGMSG("<<dst_set_voltage()\n");
    return 0;
}

int dst_set_diseqc(bktr_ptr_t bktr, P_DISEQC_VALUE pDV)
{
    DBGMSG(">>dst_set_diseqc()\n");
    
    PROTOCOL line1 = {0x0, 0x08, 0x04, 0xe0, 0x10, 0x38, 0x0, 0x0};  // LNB1
    PROTOCOL line2 = {0x0, 0x08, 0x04, 0xe0, 0x10, 0x38, 0x0, 0x0};  // LNB2  
    PROTOCOL line3 = {0x0, 0x08, 0x04, 0xe0, 0x10, 0x38, 0x0, 0x0};  // LNB3
    PROTOCOL line4 = {0x0, 0x08, 0x04, 0xe0, 0x10, 0x38, 0x0, 0x0};  // LNB4
	
	// high low band; low band bit0 = 0;
	unsigned char X4 = 0xf0;
	if (pDV->nLnb_Type == LNB_TYPE_UNIVERSAL) {
		X4 += 0x01;
	}
	// polarzation; if (H) bit1 = 0;
	if (pDV->nPolarity) {
		X4 += 0x02;
	}
	
    switch(pDV->nDiseqc_Port)
    {
    case 1:
        X4 &= 0xf3;			//0xf0,0xf1,0xf2,0xf3;
        line1[6] = X4;
        if (dstPutCommand(bktr, &line1) != 0)
            return -1;
        break;
    case 2:
        X4 += 0x04;			//0xf4,0xf5,0xf6,0xf7;
        line2[6] = X4;if (dstPutCommand(bktr, &line2) != 0)
            return -1;
        break;
    case 3:
        X4 += 0x08;			//0xf8,0xf9,0xfa,0xfb;
        line3[6] = X4;
        if (dstPutCommand(bktr, &line3) != 0)
            return -1;
        break;
    case 4:
        X4 += 0x0c;			//0xfc,0xfd,0xfe,0xff;
        line4[6] = X4;
        if (dstPutCommand(bktr, &line4) != 0)
            return -1;
        break;
    default:
        DBGMSG("n Set error! (n=1~4)\n");
        return -1;
    }
        
    DBGMSG("<<dst_set_diseqc()\n");
    return 0;
}

int dst_get_signal(bktr_ptr_t bktr, P_SIGNAL_VALUE pSV)
{
    DBGMSG(">>dst_get_signal()\n");
    
    PROTOCOL comm = {0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfb};
    PROTOCOL result = {0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfb};
    unsigned int Temp;
    
    if (dstPutCommand(bktr, &comm) != 0)
        return -1;    
    dstCommandReply(bktr, &result);    
    
    Temp = result[6];
	if (Temp & 0x10) {
		pSV->nLockFlag = 0;
	}
	else {
		pSV->nLockFlag = 1;
	}
    
    switch (bktr->card_type) {
    case 1:
    case 2:
    case 3:
    case 4:
        {
        Temp = result[2];
    	Temp = (Temp << 8) | result[3];
    
    	if (pSV->nLockFlag) {
    		if ((Temp >= 0) && (Temp <= 0x1200)) {
    			pSV->nDevSigQuality = 98;
    		}
    		else if ((Temp > 0x1200) && (Temp <= 0x1900)) {
    			double x = (double)(Temp - 0x1201) / (double)((0x1900 - 0x1201) / 57);
    			pSV->nDevSigQuality = 40 + (58 - (int)x);
    		}
    		else if ((Temp > 0x1900) && (Temp <= 0x3500)) {
    			double x = (double)(Temp - 0x1901) / (double)((0x3500 - 0x1901) / 38);
    			pSV->nDevSigQuality = 39 - (int)x;
    		}
    		else {
    			pSV->nDevSigQuality = 0;
    		}
    	}
    	else {
    		pSV->nDevSigQuality = 0;
    	}
    
    	Temp = result[5];
    	if (pSV->nLockFlag) {
    		pSV->nDevSigStrength = Temp;
    	}
    	else {
    		if (Temp > 80) {
    			pSV->nDevSigStrength = 80;
    		}
    		else {
    			pSV->nDevSigStrength = Temp;
    		}
    	}
	    }
	    break;
	case 5:
		// return data: 0x00,0x05,ErrorCounter_h,ErrorCounter_l,Strength_h,Strength_l,noise,Checksum;
		{
		Temp = result[4];
		if ((Temp & 0x04) == 0)
			pSV->nDevSigStrength = 8;
		else {
			Temp = result[5] + ((Temp & 0x03) << 8);
			double x = 100 * Temp;
			x = x / 1024;
			if ((int)x >= 90)
				pSV->nDevSigStrength = 10;
			else if ((int)x <= 15)
				pSV->nDevSigStrength = 90;
			else
				pSV->nDevSigStrength = 100 - (int)x;
		}

		Temp = result[4];
		if ((Temp & 0x80) == 0x80) {
			Temp = result[2];
			Temp = result[3] + (Temp << 8);
			if (Temp == 0) {
				if (result[6] == 0)
					pSV->nDevSigQuality = 98;
				else
					pSV->nDevSigQuality = 90;
			}
			else if (Temp < 0xf0) {
				double x = 40 + ((0x00f0 - Temp) * 20) / 0xf0;
				pSV->nDevSigQuality = (int)x;
			}
			else if (Temp < 0xf00) {
				double x = 30 + ((0x0f00 - Temp) * 10) / 0xf00;
				pSV->nDevSigQuality = (int)x;
			}
			else {
				double x = 15 + ((0xffff - Temp) * 15) / 0xffff;
				pSV->nDevSigQuality = (int)x;
			}
		}
		else {
			pSV->nDevSigQuality = 0;
		}
		}
		break;
	case 6:
	    {
		Temp = result[3];
		pSV->nDevSigQuality = Temp;
		Temp = result[5];
		pSV->nDevSigStrength = Temp;
	    }
	    break;
	case 11:
	    {
//		data1---------0x00
//		data2---------0x00 ʾû or 0x05 ̨
//		data3---------Quality(H)
//		data4---------Quality(L)
//		data5---------Strength(H)
//		data6---------Strength(L)
		Temp = result[1];
		if (Temp) {
			pSV->nLockFlag = 1;
		}
		else {
			pSV->nLockFlag = 0;
		}

		Temp = result[3];
		pSV->nDevSigQuality = Temp;
		Temp = result[5];
		pSV->nDevSigStrength = Temp;
	    }	
		break;
	}
    
    //TRYMSG("Device signal: LockFlag=%d,  strength=%d,  quality=%d\n", 
    //         pSV->nLockFlag, pSV->nDevSigStrength, pSV->nDevSigQuality);    
    DBGMSG("<<dst_get_signal()\n");
    return 0;
}        

int dst_set_lnb(bktr_ptr_t bktr, P_LNB_INFO pLnbInfo)
{
    
    DBGMSG(">>dst_set_lnb() \n");    
                    //{0x0, 0x09, 22K, Tone/Data Burst, power, 0x00, 0x00, 0x0}
    PROTOCOL comm = {0x0, 0x09, 0xff, 0xff, 0xff, 0x00, 0x00, 0x0}; 
    
    if (pLnbInfo->Power_Enable)
        comm[4] = 0x01;
        
    if (pLnbInfo->K22_Enable) {
        comm[2] = 0x02;
        comm[4] = 0x01;
    }    
    
    if (pLnbInfo->Tone_Burst_Enable) {
        comm[3] = 0x00;
        comm[4] = 0x01;
    }    
    else if (pLnbInfo->Data_Burst_Enable) {
        comm[2] = 0x02;
        comm[4] = 0x01;
    }    
    
    if (dstPutCommand(bktr, &comm) != 0) 
        return -1;
    
    DBGMSG("<<dst_set_lnb()\n");
    return 0;
}    

int dst_set_diseqc_command(bktr_ptr_t bktr, P_DISEQC_COMMAND pDiscComm)
{
    DBGMSG(">>dst_set_diseqc_command()\n");
                  //{0x0, 0x08, BitCount, Frame, Address, Command, Data1, Data2}; 
    PROTOCOL comm = {0x0, 0x08, 0x00, 0x00, 0x00, 0x00, 0x0, 0x0};
    
    switch (pDiscComm->BitCount) {
	case 3:
	    comm[2] = pDiscComm->BitCount;
	    comm[3] = pDiscComm->Frame;
	    comm[4] = pDiscComm->Address;
	    comm[5] = pDiscComm->Command;
	    comm[6] = 0x0;
		break;
	case 4:
	    comm[2] = pDiscComm->BitCount;
	    comm[3] = pDiscComm->Frame;
	    comm[4] = pDiscComm->Address;
	    comm[5] = pDiscComm->Command;
	    comm[6] = pDiscComm->Data1;
	    break;
	case 5:
	    comm[2] = pDiscComm->Frame;
	    comm[3] = pDiscComm->Address;
	    comm[4] = pDiscComm->Command;
	    comm[5] = pDiscComm->Data1;
	    comm[6] = pDiscComm->Data2;
	    break;
	default:	    
	    return -1;
	    break;
	}
	
    if (dstPutCommand(bktr, &comm) != 0)
        return -1;        
        
    DBGMSG("<<dst_set_diseqc_command()\n");
    return 0;    
}
    
///////////////////////////////////////////////////////////////////
