#define SUCCESS 0
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/types.h>
#include <linux/tty.h>
#include <linux/delay.h>
#define vremap ioremap
#define vfree iounmap
#include "twinhan_linux.h"
#include "ioctl_twinhan.h"
#include "brktree_reg.h"
#include "twinhan_common.h"
#include "dmaBuffer.h"
#include "dstnet.h"

#define PCI_VENDOR_ID_BROOKTREE 0x109e
#define PCI_DEVICE_ID_BT848     0x350
#define  TRITON_PCON	           0x50
#define  TRITON_BUS_CONCURRENCY   (1<<0)
#define  TRITON_STREAMING	  (1<<1)
#define  TRITON_WRITE_BURST	  (1<<2)
#define  TRITON_PEER_CONCURRENCY  (1<<3)
#ifndef BROOKTREE_DEF_LATENCY_VALUE
#define BROOKTREE_DEF_LATENCY_VALUE	10
#endif
#define UNIT(x)		((x) & 0x0f)
#define DEV(x)		((x >> 4) & 0x0f)
#define VIDEO_DEV	0x00
#define TUNER_DEV	0x01
static unsigned int remap = 0;
static int major = CDEV_MAJOR;
extern int printk(const char* fmt, ...);
	
static int GetDemuxData(bktr_ptr_t bktr, unsigned long count, int unit);

static void handle_chipset(void)
{
  DBGMSG(">>handle_chipset()\n");
        
  int index;  
  for (index = 0; index < 8; index++){
    unsigned char bus, devfn;
    unchar b, bo;
    if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82441,index, &bus, &devfn))
      pcibios_read_config_byte(bus, devfn, 0x53, &b);
    if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437, index, &bus, &devfn)) {
        pcibios_read_config_byte(bus, devfn, TRITON_PCON, &b);
        bo=b;
        if(!(b & TRITON_BUS_CONCURRENCY))  b |= TRITON_BUS_CONCURRENCY;
        if(b & TRITON_PEER_CONCURRENCY)    b &= ~TRITON_PEER_CONCURRENCY;
        if(!(b & TRITON_STREAMING))   b |=  TRITON_STREAMING;
        if (b!=bo)  pcibios_write_config_byte(bus, devfn, TRITON_PCON, b);
    }
  }
  
  DBGMSG("<<handle_chipset()\n");
}

static int bt848_ioctl(struct inode *ino, struct file* file, unsigned int cmd, unsigned long arg)
{
  DBGMSG(">>bt848_ioctl()\n");
        
  unsigned int minor = MINOR(ino->i_rdev);
  int unit = UNIT(minor);
  bktr_ptr_t bktr;
  int ret;
  if (unit >= brooktree_cnt) return -ENODEV;
  bktr = &brooktree[unit];
  if (bktr->bigbuf == 0) 
    return -ENOMEM;
  ret = dst_ioctl( bktr, unit, cmd, (caddr_t)arg, (struct proc *)current );
  
  DBGMSG("<<bt848_ioctl()\n");
  return ret;
}
static int bt848_open( struct inode* ino, struct file* filep)
{
  DBGMSG(">>bt848_open()\n");
  
  unsigned int minor = MINOR(ino->i_rdev);
  int unit = UNIT(minor);
  bktr_ptr_t bktr;
  if (unit >= brooktree_cnt) 
    return -ENODEV;
  bktr = &brooktree[unit];
  if (!(bktr->flags & DST_INITIALIZED)) {
    TRYMSG("bt848_open() => !(bktr->flags & DST_INITIALIZED) \n");
    return -ENXIO;
  }  
  MOD_INC_USE_COUNT;  
  
  DBGMSG("<<bt848_open()\n");
  return 0;
}

static int bt848_close( struct inode* ino, struct file* filep)
{
  DBGMSG(">>bt848_close()\n");      
  
  unsigned int minor = MINOR(ino->i_rdev);
  int unit = UNIT(minor);
  bktr_ptr_t bktr;
  
  if (unit >= brooktree_cnt)
    return -1;
  bktr = &brooktree[unit];
  dst_close(bktr);
  MOD_DEC_USE_COUNT;
  
  DBGMSG("<<bt848_close()\n");
  return 0;
}

int Init_Demux(bktr_ptr_t bktr, P_DEMULTIPLEXER_VALUE pDMV, int unit)
{
  DBGMSG(">>Init_Demux()=> pDMV->PID=%ld, DMV->StreamType=%d \n", pDMV->PID, pDMV->StreamType);
  
  //init value for GetDemuxData()   
  DMA_BUFFER_INFO     dmaBufferInfo;
  get_dmaBufferInfo(bktr, &dmaBufferInfo);       
  read_stru[unit].ReadBuf.Pos = dmaBufferInfo.currentWrite;
  read_stru[unit].ReadBuf.Len = 0;
  //memset(read_stru[unit].ReadBuf.pBuf, 0x11, READ_BUF_SIZE); //for debug
  read_stru[unit].ReadPosition = 0;
  read_stru[unit].StartBitEnable = 0;
  read_stru[unit].AStartBitEnable = 0;
  read_stru[unit].VStartBitEnable = 0;
  read_stru[unit].TempBufferPosition = 0;
  
  read_stru[unit].DMV.PID  = pDMV->PID;
  read_stru[unit].DMV.vPID = pDMV->vPID;  
  read_stru[unit].DMV.aPID = pDMV->aPID;  
  read_stru[unit].DMV.StreamType = pDMV->StreamType;
    
  read_stru[unit].TPFData.SessionLen = 0;	
  read_stru[unit].TPFData.TSSize = bktr->TS_Size;
  
  DBGMSG("<<Init_Demux() bktr->TS_Size=%d\n", bktr->TS_Size);  
  return 0;
}
    
static int bt848_read(struct file *filep, char *buf, size_t count, loff_t *offset)
{
    DBGMSG(">>bt848_read() => count=%d \n", count);
    void *pTemp = 0;
    struct inode *ino = filep->f_dentry->d_inode;
    unsigned int minor = MINOR(ino->i_rdev);
    int unit = UNIT(minor);
    bktr_ptr_t bktr;
    if (unit >= brooktree_cnt)
        return -ENXIO;
    bktr = &brooktree[unit];
    
    if (read_stru[unit].bRead_Enable)
    {
        ERRMSG("There is someone recording now... you must wait!!\n");
        return -EFAULT;
    }
    read_stru[unit].bRead_Enable = 1;    
        
    if (count > (READ_BUF_SIZE - READ_BUF_OFFSET_SIZE)) { //required data is larger than our buffer
        ERRMSG(">>bt848_read() => count > READ_BUF_SIZE !!\n");
        read_stru[unit].bRead_Enable = 0; 
        return -EFAULT;
    }    

    if (count > read_stru[unit].ReadBuf.Len) { 
        //Move the rest data to the head of ReadBuf
        if (read_stru[unit].ReadBuf.Len >= 0) {
            pTemp = (void*)(read_stru[unit].ReadBuf.pBuf + read_stru[unit].ReadBuf.Pos);
            memcpy((void*)read_stru[unit].ReadBuf.pBuf, pTemp, read_stru[unit].ReadBuf.Len);
            read_stru[unit].ReadBuf.Pos = 0;
        }
            
        if (GetDemuxData(bktr, count, unit)) {
            read_stru[unit].bRead_Enable = 0; 
            return -EFAULT;
        }
    } 
        
    //Move the rest data to the head of ReadBuf
    if ((read_stru[unit].ReadBuf.Pos + count) > READ_BUF_SIZE) { 
        pTemp = (void*)(read_stru[unit].ReadBuf.pBuf + read_stru[unit].ReadBuf.Pos);
        memcpy((void*)read_stru[unit].ReadBuf.pBuf, pTemp, read_stru[unit].ReadBuf.Len);
        read_stru[unit].ReadBuf.Pos = 0;
    }    

    // copy data to user
    if (copy_to_user(buf, (void*)(read_stru[unit].ReadBuf.pBuf+read_stru[unit].ReadBuf.Pos), count))
    {
      ERRMSG(">>bt848_read() => copy_to_user fail - 1!!\n");
      read_stru[unit].bRead_Enable = 0; 
      return -EFAULT;
    }     
    read_stru[unit].ReadBuf.Pos += count;
    read_stru[unit].ReadBuf.Len -= count;
    read_stru[unit].bRead_Enable = 0; 
    
    DBGMSG("<<bt848_read()\n");
    return count;   
}    

static int bt848_mmap(struct file *filep, struct vm_area_struct *vma)
{
  DBGMSG(">>bt848_mmap()\n");      
        
  struct inode *ino = filep->f_dentry->d_inode;
  unsigned int minor = MINOR(ino->i_rdev);
  int unit = UNIT(minor);
  bktr_ptr_t bktr;
  if (unit >= brooktree_cnt) return -ENXIO;
  bktr = &brooktree[unit];
//ERRMSG(">>bt848_mmap()=> bktr->bigbuf=%08x,  vma->vm_pgoff=%08x\n", (u_long)bktr->bigbuf, (u_long)vma->vm_pgoff);    
  vma->vm_pgoff += (u_long)bktr->bigbuf;
  if (vma->vm_pgoff & ~PAGE_MASK)    return -ENXIO;
  if (remap_page_range(vma, vma->vm_start, vtophys((vm_offset_t)vma->vm_pgoff),
    vma->vm_end - vma->vm_start, vma->vm_page_prot))
    return -EAGAIN;
  
  DBGMSG("<<bt848_mmap()\n");      
  return 0;
}

static struct file_operations bt848_fops = {
        NULL,NULL,bt848_read,NULL,NULL,NULL,bt848_ioctl,bt848_mmap,bt848_open,NULL,bt848_close,
	NULL,NULL,NULL,NULL,NULL, NULL
};

static void bt848_irq(int irq, void *dev_id, struct pt_regs * regs)
{
  //DBGMSG(">>bt848_irq()\n");
  
  bktr_ptr_t bktr = (bktr_ptr_t)dev_id;
  bktr_intr(bktr);
  
  //DBGMSG("<<bt848_irq()\n");
}

static int  find_bt848(void)
{
  DBGMSG(">>find_bt848()\n");
  
  short pci_index;    
  unchar command, latency;
  int result;
  unchar bus, devfn;
  bktr_ptr_t bktr;
  short id;
  brooktree_cnt = 0;
  if (!pcibios_present())   return 0;
  for (pci_index = 0; (!pcibios_find_device(PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT878,
			     pci_index, &bus, &devfn) || !pcibios_find_device(PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT879,
			     pci_index, &bus, &devfn));  ++pci_index)
  {
        bktr = &brooktree[brooktree_cnt];
        bktr->bus = bus;
        bktr->devfn = devfn;
        bktr->base = NULL;
        pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_INTERRUPT_LINE, &bktr->irq);
        pcibios_read_config_dword(bktr->bus, bktr->devfn, PCI_BASE_ADDRESS_0,&bktr->phys_base);
        if (remap && (!brooktree_cnt))        {
           remap <<= 20;
           remap &= PCI_BASE_ADDRESS_MEM_MASK;
           remap |= bktr->phys_base & (~PCI_BASE_ADDRESS_MEM_MASK);
           pcibios_write_config_dword(bktr->bus, bktr->devfn, PCI_BASE_ADDRESS_0,remap);
           pcibios_read_config_dword(bktr->bus, bktr->devfn, PCI_BASE_ADDRESS_0,&bktr->phys_base);
        }
        bktr->phys_base &= PCI_BASE_ADDRESS_MEM_MASK;
        pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_CLASS_REVISION,&bktr->revision);
        bktr->base = vremap(bktr->phys_base, 0x1000);
        result = request_irq(bktr->irq, bt848_irq, SA_SHIRQ | SA_INTERRUPT,"twinhan",(void *)bktr);
        if (result == -EINVAL)  return -EINVAL;
        if (result == -EBUSY)   return result;
        if (result < 0)         return result;
        pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_COMMAND, &command);
        command |= PCI_COMMAND_MASTER;
        pcibios_write_config_byte(bktr->bus, bktr->devfn, PCI_COMMAND, command);
        pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_COMMAND, &command);
        if (!(command & PCI_COMMAND_MASTER))   return -1;

        pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_LATENCY_TIMER, &latency);
        latency = 32;
        if (!latency) {
                latency = BROOKTREE_DEF_LATENCY_VALUE;
                pcibios_write_config_byte(bktr->bus, bktr->devfn, PCI_LATENCY_TIMER, latency);
        }
        brooktree_cnt++;
        pcibios_read_config_word(bktr->bus, bktr->devfn, PCI_DEVICE_ID, &id);
        if (id >= 878){
                pcibios_read_config_byte(bktr->bus, bktr->devfn, BT878_DEVCTRL,&command);
                command |= BT878_EN_TBFX;
                pcibios_write_config_byte (bktr->bus, bktr->devfn, BT878_DEVCTRL,command);
                pcibios_read_config_byte (bktr->bus, bktr->devfn, BT878_DEVCTRL,&command);
        }
        {
                bktr->f1bus = bktr->bus;
                bktr->f1devfn = bktr->devfn ;
                bktr->f1devfn ++;
                bktr->f1base = NULL;
                pcibios_read_config_byte(bktr->f1bus, bktr->f1devfn,PCI_INTERRUPT_LINE, &bktr->f1irq);
                pcibios_read_config_dword(bktr->f1bus, bktr->f1devfn, PCI_BASE_ADDRESS_0, &bktr->f1phys_base);
                bktr->f1phys_base &= PCI_BASE_ADDRESS_MEM_MASK;
                pcibios_read_config_byte(bktr->f1bus, bktr->f1devfn, PCI_CLASS_REVISION, &bktr->f1revision);
                bktr->f1base = vremap(bktr->f1phys_base, 0x1000);
                pcibios_read_config_byte(bktr->f1bus, bktr->f1devfn, PCI_COMMAND, &command);
                command |= PCI_COMMAND_MASTER;
                pcibios_write_config_byte(bktr->f1bus, bktr->f1devfn, PCI_COMMAND, command);
                pcibios_read_config_byte(bktr->f1bus, bktr->f1devfn, PCI_COMMAND, &command);
                if (!(command & PCI_COMMAND_MASTER))  return -1;
                pcibios_read_config_byte(bktr->f1bus, bktr->f1devfn, PCI_LATENCY_TIMER, &latency);
                latency = 32;
                if (!latency){
                        latency = BROOKTREE_DEF_LATENCY_VALUE;
                        pcibios_write_config_byte(bktr->f1bus, bktr->f1devfn, PCI_LATENCY_TIMER, latency);
                }
        }
    }
   
   DBGMSG("<<find_bt848()\n"); 
   return brooktree_cnt;
}
  
vm_offset_t get_bktr_mem( int unit, unsigned size )
{
  DBGMSG(">>get_bktr_mem()\n");
  
  vm_offset_t	addr = 0;
  addr = kmalloc(size, GFP_KERNEL);
  
  DBGMSG("<<get_bktr_mem()\n");
  return( addr );
}

static int init_bt848(int unit)
{
  DBGMSG(">>init_bt848()\n");
  
  bktr_ptr_t	bktr;
  bt848_ptr_t	bt848;
  vm_offset_t	buf;
  bktr = &brooktree[unit];
  bktr->dma_prog = get_bktr_mem(unit, DMA_PROG_ALLOC);
//TRYMSG("<<register_chrdev() unit=%d\n", unit);  
  if (bktr->dma_prog == 0 )
    return -1;
  if ( BROOKTREE_ALLOC ) {
      buf = bigphysarea_alloc(BROOKTREE_ALLOC);
      if (buf == 0) {
        ERRMSG(">>init_bt848() => bigphysarea_alloc fail !!\n");
        return -1;
      }  
  }    
  else 
      buf = 0;       
  bktr->bigbuf = buf;
  bktr->alloc_pages = BROOKTREE_ALLOC_PAGES;
  if (buf != 0)  {
    memset((caddr_t) buf, 0, BROOKTREE_ALLOC);
    buf = (char *)vtophys(buf);
    bktr->flags = DST_INITIALIZED;
    bktr->dma_prog_loaded = FALSE;
    bt848 = bktr->base;
    bt848->int_mask = ALL_INTS_DISABLED;
    bt848->gpio_dma_ctl = FIFO_RISC_DISABLED;
  }    
  bktr->tflags = TUNER_INITALIZED;
  bktr->flag22K = 0;
  bktr->flag1318V = 1;
  check_card_version(bktr);
  
  read_stru[unit].ReadBuf.pBuf = (void*)__get_dma_pages(GFP_KERNEL, ReadBuf_PAGE_ORDER);;
  if (read_stru[unit].ReadBuf.pBuf == 0 ) {
    ERRMSG("alloc ReadBuf.pRead_Buf fail !!\n");
    return -1;
  }
    
  DBGMSG("<<init_bt848()\n");
  return 0;
}
static void release_bt848 (void)
{
  DBGMSG(">>release_bt848()\n");
  
  unchar command;
  int i;
  bktr_ptr_t bktr;
  bt848_ptr_t bt848;
  for (i = 0; i < brooktree_cnt; i++)  {
    bktr = &brooktree[i];
    bt848 = bktr->base;
    bt848->int_mask = ALL_INTS_DISABLED;
    bt848->int_stat = ALL_INTS_CLEARED;
    bt848->gpio_out_en = 0;
    pcibios_read_config_byte(bktr->bus, bktr->devfn, PCI_COMMAND, &command);
    command |= PCI_COMMAND_MASTER;
    pcibios_write_config_byte(bktr->bus, bktr->devfn, PCI_COMMAND, command);
    if (bktr->bigbuf) bigphysarea_free((char *)bktr->bigbuf, bktr->alloc_pages);
    if (bktr->dma_prog)  kfree((char *)bktr->dma_prog);
    free_irq(bktr->irq, bktr);
    if (bktr->base)   vfree((char *)bktr->base);
    if (read_stru[i].ReadBuf.pBuf) free_pages((unsigned long)read_stru[i].ReadBuf.pBuf, ReadBuf_PAGE_ORDER);
  }
  
  DBGMSG("<<release_bt848()\n");
}
int init_module( void)
{
  DBGMSG(">>init_module()\n");
        
  int i;
   
  handle_chipset();
  if (find_bt848() < 0) {
    release_bt848();
    release_dmaBuffer();
    return -EIO;
  }  
  
  if (init_dmaBuffer()) {
    release_bt848();
    release_dmaBuffer();
    return -EIO;
  } 
  
  RETMSG(">>init_module() brooktree_cnt=%d\n", brooktree_cnt);
  for (i = 0; i < brooktree_cnt; i++) {
    if (init_bt848(i) < 0){
      ERRMSG("<<init_bt848() fail !! \n");  
      release_bt848();
      release_dmaBuffer();
      return -EIO;
    }
  }  
  
  if (register_chrdev(major, "twinhan", &bt848_fops)){
    ERRMSG("<<register_chrdev() fail !!  \n");
    release_bt848();
    release_dmaBuffer();
    return -EIO;
  }
  //dst_net_init();
  
  DBGMSG("<<init_module()\n");
  return 0;
}

void cleanup_module(void)
{
  DBGMSG(">>cleanup_module()\n");
  
  release_bt848();
  if (unregister_chrdev(major, "twinhan") != 0)
    ERRMSG("twinhan: error: could not remove driver\n");
  else
    DBGMSG("twinhan: **** twinhan Driver Removed *****\n");
  release_dmaBuffer();
  //dst_net_release();
  
  DBGMSG("<<cleanup_module()\n");
}

#define MAXTRYNUM 50000
static int GetDemuxData(bktr_ptr_t bktr, unsigned long count, int unit)
{ 
    DBGMSG(">>GetDemuxData()  => unit=%d  ReadBuf.Len=%ld\n", unit, ReadBuf.Len);
    
    DMA_BUFFER_INFO     dmaBufferInfo;    
    unsigned char       *pDMABufferHead = (unsigned char*)bktr->bigbuf;        
    unsigned long       ValidBufSize = 0;
    unsigned long       WritePosition = 0;    
    unsigned char       *pDMABufReadIdx = 0;
    int		            SyncFindFlag = 0;	
	unsigned long	    CurrentPosition = 0;	
	unsigned char	    Temp = 0;		
	int         		ByteCnt = 0;        
	read_stru[unit].TPFData.SessionLen = 0;	
	int nTryNum = 0;

    while ((count > read_stru[unit].ReadBuf.Len) && (read_stru[unit].ReadBuf.Len < (READ_BUF_SIZE - READ_BUF_OFFSET_SIZE)) && (nTryNum < MAXTRYNUM)) {
        get_dmaBufferInfo(bktr, &dmaBufferInfo);       
        WritePosition = dmaBufferInfo.currentWrite;
        CurrentPosition = 0;
        if (read_stru[unit].ReadPosition <= WritePosition) {
            ValidBufSize = WritePosition - read_stru[unit].ReadPosition;
    	}
    	else {
    		ValidBufSize = DMABUFFERLEN_2 - read_stru[unit].ReadPosition;
    	}
    	
    	if (ValidBufSize == 0) {
    	    current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ*10/100); //60ms	
            nTryNum++;
			continue;
		}
		
        if (ValidBufSize) {
            while (CurrentPosition < ValidBufSize) {
				if (read_stru[unit].TempBufferPosition) {
					pDMABufReadIdx = (unsigned char *)pDMABufferHead + read_stru[unit].ReadPosition;
					memcpy((void *)(read_stru[unit].TempBuffer + read_stru[unit].TempBufferPosition),
						   (void *)(pDMABufReadIdx + CurrentPosition),
						   read_stru[unit].TPFData.TSSize - read_stru[unit].TempBufferPosition);
					pDMABufReadIdx = read_stru[unit].TempBuffer;
					pDMABufReadIdx++;
					SyncFindFlag = 1;
				}
				else {
					pDMABufReadIdx = (unsigned char *)pDMABufferHead + read_stru[unit].ReadPosition + CurrentPosition;
					SyncFindFlag = 0;
				}				
				while (!SyncFindFlag) {
					if (CurrentPosition >= ValidBufSize) {
						break;
					}
					Temp = *pDMABufReadIdx++;
					CurrentPosition++;
					if (Temp == 0x47) {		    					    
						if ((CurrentPosition + read_stru[unit].TPFData.TSSize - 1) < ValidBufSize) {
							if (*(pDMABufReadIdx + read_stru[unit].TPFData.TSSize - 1) == 0x47) {
								SyncFindFlag = 1;
							}
						}
						else {
							SyncFindFlag =1;							
						}
					}
				}

				if ((CurrentPosition >= ValidBufSize) && !SyncFindFlag) {				    
					continue;					
				}				
				ByteCnt = 187;
				if ((CurrentPosition + read_stru[unit].TPFData.TSSize - 1) <= ValidBufSize) {
					Temp = *pDMABufReadIdx++;
					read_stru[unit].TPFData.payload_unit_start_indicator = (Temp >> 6) & 0x01;
					read_stru[unit].TPFData.pid	= ((Temp & 0x1f) << 8) | *pDMABufReadIdx++;        			
					read_stru[unit].TPFData.transport_scrambling_control = (*pDMABufReadIdx & 0xc0) >> 6;
					ByteCnt -= 2;

                    if((read_stru[unit].TPFData.transport_scrambling_control==2) ||(read_stru[unit].TPFData.transport_scrambling_control==3))
				        goto NextLoop;

					Temp = *pDMABufReadIdx++;
					ByteCnt--;
					read_stru[unit].TPFData.adaptation_field_control = (Temp >> 4) & 0x3;
					if (read_stru[unit].TPFData.adaptation_field_control > 1){
						read_stru[unit].TPFData.adaptation_field_length = *pDMABufReadIdx++;
						ByteCnt--;
						if (read_stru[unit].TPFData.adaptation_field_length >= ByteCnt){
							CurrentPosition += read_stru[unit].TPFData.TSSize - 1;
							read_stru[unit].TempBufferPosition = 0;
							continue;
						}
						pDMABufReadIdx += read_stru[unit].TPFData.adaptation_field_length;
						ByteCnt -= read_stru[unit].TPFData.adaptation_field_length;
					}	
					
					if (TH_ALL_PID_TS_PACKET == read_stru[unit].DMV.StreamType) {
					    pDMABufReadIdx -= 188 - ByteCnt;
						ByteCnt = 188;						
					    memcpy((void *)(read_stru[unit].TPFData.pSession), pDMABufReadIdx, ByteCnt);
					    read_stru[unit].TPFData.SessionLen = ByteCnt;
						    
						goto NextLoop;  //Get next TS packet
					}	
					
					if (TH_VA_TS_PACKET == read_stru[unit].DMV.StreamType) {
                        if ((read_stru[unit].TPFData.pid == read_stru[unit].DMV.vPID) || (read_stru[unit].TPFData.pid == read_stru[unit].DMV.aPID)) {					    
    					    pDMABufReadIdx -= 188 - ByteCnt;
    						ByteCnt = 188;    						
    					    memcpy((void *)(read_stru[unit].TPFData.pSession), pDMABufReadIdx, ByteCnt);
    					    read_stru[unit].TPFData.SessionLen = ByteCnt;    						    
    						goto NextLoop;  //Get next TS packet
    					}	
    					else
    					    goto NextLoop;  //Get next TS packet
					}
					
					/*if (TH_PID_PS_PACKET == read_stru[unit].DMV.StreamType) {
					    if ((read_stru[unit].TPFData.adaptation_field_control==1) || (read_stru[unit].TPFData.adaptation_field_control==3)) {
					        //Get Video PES PACKET 
					        if(read_stru[unit].TPFData.pid == read_stru[unit].DMV.vPID) {
        						if (read_stru[unit].TPFData.payload_unit_start_indicator) {
        							read_stru[unit].VStartBitEnable = 1;
        						}        
        						if (read_stru[unit].VStartBitEnable) {
        							read_stru[unit].TPFData.SessionLen = ByteCnt;        							
        						}
        					}
        					//Get Audio PES PACKET 
					        if(read_stru[unit].TPFData.pid == read_stru[unit].DMV.aPID) {
        						if (read_stru[unit].TPFData.payload_unit_start_indicator) {
        							read_stru[unit].AStartBitEnable = 1;
        						}        
        						if (read_stru[unit].AStartBitEnable) {
        							read_stru[unit].TPFData.SessionLen = ByteCnt;
        						}
        					}	
					    }
					    goto NextLoop;  //Get next TS packet
					}*/	   
										 
					if (((TH_PID_TS_PACKET == read_stru[unit].DMV.StreamType) || (TH_PID_PES_PACKET == read_stru[unit].DMV.StreamType))
					    && (read_stru[unit].TPFData.pid != read_stru[unit].DMV.PID))
                            goto NextLoop;  //Get next TS packet      
                            
					if ((read_stru[unit].TPFData.adaptation_field_control==1) || (read_stru[unit].TPFData.adaptation_field_control==3)) {
					    //Get TS PACKET
						if (TH_PID_TS_PACKET == read_stru[unit].DMV.StreamType) {  
							pDMABufReadIdx -= 188 - ByteCnt;
							ByteCnt = 188;
							
							if (read_stru[unit].TPFData.payload_unit_start_indicator) {
								read_stru[unit].StartBitEnable = 1;
							} 
							
							if (read_stru[unit].StartBitEnable) {
							    memcpy((void *)(read_stru[unit].TPFData.pSession), pDMABufReadIdx, ByteCnt);
							    read_stru[unit].TPFData.SessionLen = ByteCnt;
							}
							goto NextLoop;  //Get next TS packet
						}          
						
                        //Get PES PACKET 
						if (read_stru[unit].TPFData.payload_unit_start_indicator) {
							read_stru[unit].StartBitEnable = 1;
						}

						if (read_stru[unit].StartBitEnable) {			    
							memcpy((void *)(read_stru[unit].TPFData.pSession), pDMABufReadIdx, ByteCnt);
							read_stru[unit].TPFData.SessionLen = ByteCnt;
        				}	
					}
					
NextLoop:
					if (read_stru[unit].TempBufferPosition) {
						CurrentPosition += read_stru[unit].TPFData.TSSize - read_stru[unit].TempBufferPosition;
						read_stru[unit].TempBufferPosition = 0;
					}
					else {
						CurrentPosition += read_stru[unit].TPFData.TSSize - 1;
					}
				}
				else {
					pDMABufReadIdx--;
					CurrentPosition--;
					memcpy((void *)read_stru[unit].TempBuffer, (void *)pDMABufReadIdx, ValidBufSize - CurrentPosition);
					read_stru[unit].TempBufferPosition += ValidBufSize - CurrentPosition;
					CurrentPosition = ValidBufSize;
				}

				if (read_stru[unit].TPFData.SessionLen > 0) {
				    nTryNum = 0;
    				if ((read_stru[unit].ReadBuf.Len + read_stru[unit].TPFData.SessionLen) < READ_BUF_SIZE) {     				
    				    memcpy((void*)(read_stru[unit].ReadBuf.pBuf+read_stru[unit].ReadBuf.Len), (void*)read_stru[unit].TPFData.pSession, read_stru[unit].TPFData.SessionLen);
                        read_stru[unit].ReadBuf.Len += read_stru[unit].TPFData.SessionLen;
                    }
                    else { //try to increase READ_BUF_OFFSET_SIZE value
                        DBGMSG(">>GetDemuxData() => (ReadBuf.Len + TPFData.SessionLen) > READ_BUF_SIZE \n"); 
                        return 0;
                    }
                }
                else {
                    nTryNum++;			    
                }    
                read_stru[unit].TPFData.SessionLen = 0;	
			}
			read_stru[unit].ReadPosition += ValidBufSize;
			if (read_stru[unit].ReadPosition >= DMABUFFERLEN_2) {
				read_stru[unit].ReadPosition = 0;
			}
        }
    }    
    
    if (nTryNum >= MAXTRYNUM) {
        DBGMSG(">>GetDemuxData() => (nTryNum >= MAXTRYNUM) => unit=%d \n", unit);
        return 1;
    }    
    DBGMSG("<<GetDemuxData()  \n");
    return 0;
}

//*****************************************************************************************************************//
